"""Device utilities"""

import os
import shutil
import subprocess as sp
from typing import Union

from zea import log


def check_nvidia_smi():
    """Checks whether nvidia-smi is available."""
    return shutil.which("nvidia-smi") is not None


def hide_gpus(gpu_ids=None, verbose=True):
    """Hides the specified GPUs from the system by setting the
    CUDA_VISIBLE_DEVICES environment variable.

    This can be useful when some GPUs have too little tensor cores
    to be useful for training, or when some GPUs are reserved for
    other tasks.

    Args:
        gpu_ids (list): list of GPU ids to hide.
    """
    if gpu_ids is None:
        return
    assert isinstance(gpu_ids, (int, list)), (
        f"gpu_ids must be an integer or a list of integers, not {type(gpu_ids)}"
    )
    if not isinstance(gpu_ids, list):
        gpu_ids = [gpu_ids]

    hide_gpu_ids = gpu_ids
    all_gpu_ids = list(range(len(get_gpu_memory(verbose=False))))
    keep_gpu_ids = [x for x in all_gpu_ids if x not in hide_gpu_ids]

    if len(keep_gpu_ids) == 0:
        log.warning("All GPUs are hidden. Setting CUDA_VISIBLE_DEVICES to an empty string.")
        os.environ["CUDA_VISIBLE_DEVICES"] = ""
    else:
        os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, keep_gpu_ids))
        if len(hide_gpu_ids) > 0:
            if verbose:
                print(f"Hiding GPUs {hide_gpu_ids} from the system.")


def print_gpu_memory_table(memory_free_values):
    """Prints a table of GPU memory similar to pandas DataFrame output."""
    # Header
    print("     memory")
    print("GPU        ")
    # Rows
    for idx, mem in enumerate(memory_free_values):
        print(f"{idx:<6}{mem:>7}")


def get_gpu_memory(verbose=True):
    """Retrieve memory allocation information of all gpus.

    Args:
        verbose (bool): prints output if True.

    Returns:
        memory_free_values: list of available memory for each gpu in MiB.
    """
    if not check_nvidia_smi():
        log.warning(
            "nvidia-smi is not available. Please install nvidia-utils. "
            "Cannot retrieve GPU memory. Falling back to CPU.."
        )
        return None

    def _output_to_list(x):
        return x.decode("ascii").split("\n")[:-1]

    COMMAND = "nvidia-smi --query-gpu=memory.free --format=csv"
    try:
        memory_free_info = _output_to_list(sp.check_output(COMMAND.split()))[1:]
    except Exception as e:
        print(f"An error occurred: {e}")

    memory_free_values = [int(x.split()[0]) for i, x in enumerate(memory_free_info)]

    # only show enabled devices
    if os.environ.get("CUDA_VISIBLE_DEVICES", "") != "":
        gpus = os.environ["CUDA_VISIBLE_DEVICES"]
        gpus = [int(gpu) for gpu in gpus.split(",")][: len(memory_free_values)]
        if verbose:
            # Report the number of disabled GPUs out of the total
            num_disabled_gpus = len(memory_free_values) - len(gpus)
            num_gpus = len(memory_free_values)

            print(f"{num_disabled_gpus / num_gpus} GPUs were disabled")

        memory_free_values = [memory_free_values[gpu] for gpu in gpus]

    if verbose:
        print_gpu_memory_table(memory_free_values)

    return memory_free_values


def select_gpus(available_gpu_ids, memory_free, device=None, verbose=True, hide_others=True):
    """Select GPU based on the device argument and available GPU's. This
    function does not rely on pytorch or tensorflow, and is shared between both
    frameworks.

    Hides other GPUs from the system by default by setting the
    CUDA_VISIBLE_DEVICES environment variable. Use the hide_others argument to
    disable this behavior.

    Args:
        available_gpu_ids (list): list of available GPU ids.
        memory_free (list): list of available memory for each gpu in MiB.
        device (str/int/list): GPU device(s) to select.
            - If 'cpu', use CPU. This function will be a no-op.
            - If 'gpu', select GPU based on available memory.
                Throw an error if no GPU is available.
            - If None, try to select GPU based on available memory.
                Fall back to CPU if no GPU is available.
            - If an integer or a list of integers, use the corresponding GPU(s).
                If the list contains None values (e.g. [0, None, 2]), a GPU
                will be selected based on available memory.
            - If formatted as 'cuda:xx' or 'gpu:xx', where xx is an integer,
                use the corresponding GPU(s).
            - If formatted as 'auto:xx', where xx is an integer, automatically
                select xx GPUs based on available memory. If xx is -1, use all
                available GPUs.
        verbose (bool): prints output if True.
        hide_others (bool): if True, hide other GPUs from the system by setting
            the CUDA_VISIBLE_DEVICES environment variable.

    Returns:
        gpu_ids: list of selected GPU ids. If no GPU is selected, returns an
            empty list. If a CPU is selected, returns None.
    """
    gpu_ids = []
    # Check if GPU mode is forced or if GPU should be selected based on memory
    if device == "cpu" or (device is None and not available_gpu_ids):
        print("Setting device to CPU")
        return None
    elif device == "gpu" or device == "cuda" or device is None:
        # Use None to select GPU based on available memory later
        gpu_ids = [None]
    elif isinstance(device, int) or device is None:
        gpu_ids = [device]  # Use a specific GPU if an integer is provided
    elif isinstance(device, list):
        gpu_ids = device  # Use multiple specific GPUs if a list of integers is provided
    elif isinstance(device, str):
        device = device.lower()  # Parse the device string

        if device.startswith("cuda:") or device.startswith("gpu:"):
            # Parse and use a specific GPU or all GPUs
            device_id = int(device.split(":")[1])

            if not isinstance(device_id, int):
                raise ValueError(f'Invalid device format: {device}. Expected "cuda:<gpu_id>".')
            gpu_ids = [device_id]

        elif device.startswith("auto:"):
            # Automatically select GPUs based on available memory
            num_gpus = int(device.split(":")[1])  # number of GPUs to use

            # num_gpus can be -1 which means use all available GPUs
            if verbose:
                if num_gpus == -1:
                    print("Selecting all available GPUs.")
                elif num_gpus == 0:
                    print("Not using any GPUs.")
                elif num_gpus == 1:
                    print("Selecting 1 GPU based on available memory.")
                else:
                    print(f"Selecting {num_gpus} GPUs based on available memory.")

            if not isinstance(num_gpus, int):
                raise ValueError(f'Invalid device format: {device}. Expected "auto:<num_gpus>".')
            if num_gpus == -1:
                num_gpus = len(available_gpu_ids)  # use all available GPUs
            # Create list of N None values corresponding to unassigned GPUs
            gpu_ids = num_gpus * [None]

        else:
            raise ValueError(f"Invalid device format: {device}. ")

    # Auto-select GPUs based on available memory for None values
    if None in gpu_ids:
        # Automatically select GPUs based on available memory
        sorted_gpu_ids = [
            x for x, _ in sorted(enumerate(memory_free), key=lambda x: x[1], reverse=True)
        ]

        assert len(gpu_ids) <= len(sorted_gpu_ids), (
            f"Selected more GPUs ({len(gpu_ids)}) than available ({len(sorted_gpu_ids)})"
        )

        for i, gpu in enumerate(gpu_ids):
            if gpu is None and sorted_gpu_ids[i] in available_gpu_ids:
                gpu_ids[i] = sorted_gpu_ids[i]
    else:
        bad_gpus = set(gpu_ids) - set(available_gpu_ids)
        if bad_gpus:
            raise ValueError(f"GPUs {bad_gpus} not available!!")

    if verbose:
        for gpu_id in gpu_ids:
            print(f"Selected GPU {gpu_id} with Free Memory: {memory_free[gpu_id]:.2f} MiB")

    # Hide other GPUs from the system
    if hide_others:
        hide_gpu_ids = [x for x in available_gpu_ids if x not in gpu_ids]
        hide_gpus(hide_gpu_ids, verbose=verbose)

    return gpu_ids


def get_device(device="auto:1", verbose=True, hide_others=True):
    """Sets the GPU usage by searching for available GPUs and
    selecting one or more GPUs based on the device argument.
    If CUDA is unavailable, fallback to CPU.

    Hides other GPUs from the system by default by setting the
    CUDA_VISIBLE_DEVICES environment variable. Use the hide_others argument to
    disable this behavior.

    Args:
        device (str/int/list): GPU device(s) to select. Defaults to 'auto:1'.
            - If 'cpu', use CPU.
            - If 'gpu', select GPU based on available memory.
                Throw an error if no GPU is available.
            - If None, try to select GPU based on available memory.
                Fall back to CPU if no GPU is available.
            - If an integer or a list of integers, use the corresponding
                GPU(s). If the list contains None values (e.g. [0, None, 2]), a
                GPU will be selected based on available memory.
            - If formatted as 'cuda:xx' or 'gpu:xx', where xx is an integer,
                use the corresponding GPU(s).
            - If formatted as 'auto:xx', where xx is an integer, automatically
                select xx GPUs based on available memory. If xx is -1, use all available GPUs.
        verbose (bool): prints output if True.
        hide_others (bool): if True, hide other GPUs from the system by setting
            the CUDA_VISIBLE_DEVICES environment variable.

    Returns:
        gpu_ids: list of selected GPU ids. If no GPU is selected, returns an
            empty list. If a CPU is selected, returns None.
    """

    def _cpu_case():
        os.environ["JAX_PLATFORMS"] = "cpu"  # only affects jax
        if hide_others:
            os.environ["CUDA_VISIBLE_DEVICES"] = ""
        # returns None to indicate CPU

    if device.lower() == "cpu":
        return _cpu_case()

    if verbose:
        header = "GPU settings"
        print("-" * 2 + header.center(50 - 4, "-") + "-" * 2)

    memory = get_gpu_memory(verbose=verbose)
    if memory is None:  # nvidia-smi not working, fallback to CPU
        return _cpu_case()

    gpu_ids = list(range(len(memory)))

    selected_gpu_ids = select_gpus(
        available_gpu_ids=gpu_ids,
        memory_free=memory,
        device=device,
        verbose=verbose,
        hide_others=hide_others,
    )

    if verbose:
        print("-" * 50)

    return selected_gpu_ids


def backend_cuda_available(backend):
    """Check if the selected backend is installed with CUDA support."""
    if backend == "torch":
        try:
            import torch
        except Exception:
            return False
        return torch.cuda.is_available()
    if backend == "tensorflow":
        try:
            import tensorflow as tf
        except Exception:
            return False
        return bool(tf.config.list_physical_devices("GPU"))
    if backend == "jax":
        try:
            import jax
        except Exception:
            return False
        try:
            return bool(jax.devices("gpu"))
        except Exception:
            return False
    return False


def backend_key(backend):
    """Returns cuda/gpu for the given backend"""
    if backend == "torch":
        return "cuda"
    if backend == "tensorflow":
        return "gpu"
    if backend == "jax":
        return "gpu"
    return "gpu"


def selected_gpu_ids_to_device(selected_gpu_ids, backend):
    """Convert selected GPU ids to device string."""
    if selected_gpu_ids is None or len(selected_gpu_ids) == 0:
        return "cpu"

    if len(selected_gpu_ids) > 1:
        log.warning(
            (
                "Specified multiple GPU's but this function will just return "
                f"one GPU: {selected_gpu_ids[0]}"
            )
        )

    key = backend_key(backend)
    if backend == "jax":
        # Because jax hides the other gpus, we need to set the device number to 0
        return f"{key}:0"
    else:
        return f"{key}:{selected_gpu_ids[0]}"


def set_memory_growth_tf():
    """Attempts to allocate only as much GPU memory as needed for the runtime allocations"""
    try:
        import tensorflow as tf
    except Exception:
        return

    try:
        # Currently, memory growth needs to be the same across GPUs
        for gpu in tf.config.get_visible_devices("GPU"):
            tf.config.experimental.set_memory_growth(gpu, True)
    except RuntimeError as e:
        print(e)


def init_device(
    device: Union[str, int, list] = "auto:1",
    backend: Union[str, None] = "auto",
    hide_devices: Union[int, list] = None,
    allow_preallocate: bool = True,
    verbose: bool = True,
):
    """Selects a GPU or CPU device based on the config.

    Args:
        backend (str): String indicating which backend to use. Can be
            'torch', 'tensorflow', 'jax', 'numpy', `None` or "auto".
            - When "auto", the function will select the backend based on the
            `KERAS_BACKEND` environment variable.
            - For numpy this function will return 'cpu'.
        device (str/int/list): device(s) to select.
            Examples: 'cuda:1', 'gpu:2', 'auto:-1', 'cpu', 0, or [0,1,2,3].
            For more details see: `get_device`.
        hide_devices (int/list): device(s) to hide from the system.
            Examples: 0, or [0,1,2,3]. Can be useful when some GPUs have too
            little tensor cores to be useful for training, or when some GPUs
            are reserved for other tasks. Defaults to None, in which case no
            GPUs are hidden and all are available for use.
        allow_preallocate (bool, optional): allow preallocation of memory.
            Used for jax and tensorflow.
        verbose (bool, optional): print device selection. Defaults to True.
    Returns:
        device (str/int/list): selected device(s).
    """
    if hide_devices is not None:
        hide_gpus(hide_devices)

    # Get backend from environment variable
    if backend == "auto":
        backend = os.environ.get("KERAS_BACKEND")

    if backend in ["jax", "tensorflow", "torch"]:
        selected_gpu_ids = get_device(device, verbose=verbose)
        device = selected_gpu_ids_to_device(selected_gpu_ids, backend)
    elif backend in ["numpy", "cpu"]:
        device = "cpu"
    else:
        raise ValueError(f"Unknown backend ({backend}) in config.")

    # Early exit if device is CPU
    if device == "cpu":
        return device

    # Set if jax and tensorflow should preallocate memory
    if not allow_preallocate:
        os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"
        set_memory_growth_tf()

    # Check if the selected backend is installed with CUDA support
    # -> Run this last because it will mess up the hiding of GPUs!
    if not backend_cuda_available(backend):
        device = "cpu"

    return device



"""LPIPS model for perceptual similarity.

See original code: https://github.com/richzhang/PerceptualSimilarity
As well as the paper: https://arxiv.org/abs/1801.03924

"""

import keras
from keras import ops
from keras.layers import Conv2D, Dropout, Input

from zea.internal.registry import model_registry
from zea.models.base import BaseModel
from zea.models.preset_utils import get_preset_loader, register_presets
from zea.models.presets import lpips_presets


@model_registry(name="lpips")
class LPIPS(BaseModel):
    """Learned Perceptual Image Patch Similarity (LPIPS) metric."""

    def __init__(
        self,
        net_type="vgg",
        disable_checks=False,
        **kwargs,
    ):
        """Initialize the LPIPS model.

        Exported weights using:
            https://github.com/moono/lpips-tf2.x/blob/master/example_export_script/convert_to_tensorflow.py

        Args:
            net_type (str, optional): Type of network to use. Defaults to "vgg".
            disable_checks (bool, optional): Disable input checks. This is useful to allow
                tensorflow graph mode. Defaults to False.
        """
        super().__init__(**kwargs)

        assert net_type == "vgg", "Only VGG model is supported"

        self.net = perceptual_model()
        self.lin = linear_model()
        self.disable_checks = disable_checks
        self.trainable = False  # for keras: makes the weights non-trainable

    def custom_load_weights(self, preset, **kwargs):
        """Load the weights for the VGG and linear models."""
        loader = get_preset_loader(preset)

        vgg_file = "vgg/vgg.weights.h5"
        lin_file = "lin/lin.weights.h5"
        vgg_file = loader.get_file(vgg_file)
        lin_file = loader.get_file(lin_file)

        self.net.load_weights(vgg_file, **kwargs)
        self.lin.load_weights(lin_file, **kwargs)

    @staticmethod
    def _normalize_tensor(in_feat, eps: float = 1e-8):
        """Normalize input tensor."""
        norm_factor = ops.sqrt(eps + ops.sum(in_feat**2, axis=-1, keepdims=True))
        return in_feat / norm_factor

    def call(self, inputs):
        """Compute the LPIPS metric.

        Args:
            inputs (list): List of two input images of shape [B, H, W, C] or [H, W, C].
                Images should be in the range [-1, 1].

        """
        input1, input2 = inputs

        # check input images
        if not self.disable_checks and not (self._valid_img(input1) and self._valid_img(input2)):
            raise ValueError(
                "Expected both input arguments to be normalized tensors with shape [B, H, W, C]"
                f" or [H, W, C]. Got input with shape {input1.shape} and {input2.shape} and values"
                f" in range {[ops.min(input1), ops.max(input1)]} and"
                f" {[ops.min(input2), ops.max(input2)]} when all values are expected to be in"
                " the [-1, 1] range."
            )

        # preprocess input images
        net_out1 = self.preprocess_input(input1)
        net_out2 = self.preprocess_input(input2)

        # run vgg model first
        net_out1 = self.net(net_out1)
        net_out2 = self.net(net_out2)

        # normalize
        net_out1 = [self._normalize_tensor(t) for t in net_out1]
        net_out2 = [self._normalize_tensor(t) for t in net_out2]

        # subtract
        diffs = [ops.square(t1 - t2) for t1, t2 in zip(net_out1, net_out2)]

        # run on learned linear model
        lin_out = self.lin(diffs)

        # take spatial average: list([N, 1], [N, 1], [N, 1], [N, 1], [N, 1])
        lin_out = ops.convert_to_tensor([ops.mean(t, axis=[1, 2], keepdims=False) for t in lin_out])

        # take sum of all layers: [N, 1]
        lin_out = ops.sum(lin_out, axis=0)

        # squeeze: [N, ]
        lin_out = ops.squeeze(lin_out, axis=-1)

        return lin_out

    @staticmethod
    def _valid_img(img) -> bool:
        """Check that input is a valid image to the network."""
        value_check = ops.max(img) <= 1.0 and ops.min(img) >= -1
        # singleton dim gets broadcasted to 3 RGB channels
        shape_check = ops.ndim(img) in [3, 4] and ops.shape(img)[-1] in [1, 3]
        return shape_check and value_check

    @staticmethod
    def preprocess_input(image):
        """Preprocess the input images

        Args:
            image (Tensor): Input image tensor of shape [B, H, W, C]
                and values in the range [-1, 1].
        Returns:
            Tensor: Preprocessed image tensor of shape [B, H, W, C]
                and standardized values for VGG model.
        """

        scale = ops.convert_to_tensor([0.458, 0.448, 0.450])[None, None, None, :]
        shift = ops.convert_to_tensor([-0.030, -0.088, -0.188])[None, None, None, :]
        image = (image - shift) / scale
        return image


def perceptual_model():
    """Get the VGG16 model for perceptual loss."""
    layers = [
        "block1_conv2",
        "block2_conv2",
        "block3_conv3",
        "block4_conv3",
        "block5_conv3",
    ]
    vgg16 = keras.applications.vgg16.VGG16(include_top=False, weights=None)

    vgg16_output_layers = [layer.output for layer in vgg16.layers if layer.name in layers]
    model = keras.Model(vgg16.input, vgg16_output_layers, name="perceptual_model")
    return model


def linear_model():
    """Get the linear head model for LPIPS."""
    vgg_channels = [64, 128, 256, 512, 512]
    inputs, outputs = [], []
    for ii, channel in enumerate(vgg_channels):
        name = f"lin{ii}"

        model_input = Input(shape=(None, None, channel), dtype="float32")
        model_output = Dropout(rate=0.5, dtype="float32")(model_input)
        model_output = Conv2D(
            filters=1,
            kernel_size=1,
            strides=1,
            use_bias=False,
            dtype="float32",
            data_format="channels_last",
            name=name,
        )(model_output)
        inputs.append(model_input)
        outputs.append(model_output)

    model = keras.Model(inputs=inputs, outputs=outputs, name="linear_model")
    return model


register_presets(lpips_presets, LPIPS)



"""Preset utils for zea datasets hosted on Hugging Face.

See https://huggingface.co/zea/
"""

from pathlib import Path

from huggingface_hub import hf_hub_download, list_repo_files, login
from huggingface_hub.utils import (
    EntryNotFoundError,
    HFValidationError,
    RepositoryNotFoundError,
)

from zea.internal.cache import ZEA_CACHE_DIR

HF_DATASETS_DIR = ZEA_CACHE_DIR / "huggingface" / "datasets"
HF_DATASETS_DIR.mkdir(parents=True, exist_ok=True)

HF_SCHEME = "hf"
HF_PREFIX = "hf://"


def _hf_parse_path(hf_path: str):
    """Parse hf://repo_id[/subpath] into (repo_id, subpath or None)."""
    if not hf_path.startswith(HF_PREFIX):
        raise ValueError(f"Invalid hf_path: {hf_path}. It must start with '{HF_PREFIX}'.")
    path = hf_path.removeprefix(HF_PREFIX)
    parts = path.split("/")
    repo_id = "/".join(parts[:2])
    subpath = "/".join(parts[2:]) if len(parts) > 2 else None
    return repo_id, subpath


def _hf_list_files(repo_id):
    try:
        files = list_repo_files(repo_id, repo_type="dataset")
    except (RepositoryNotFoundError, HFValidationError, EntryNotFoundError):
        login(new_session=False)
        files = list_repo_files(repo_id, repo_type="dataset")
    return files


def _hf_download(repo_id, filename, cache_dir=HF_DATASETS_DIR):
    return hf_hub_download(
        repo_id=repo_id,
        filename=filename,
        cache_dir=cache_dir,
        repo_type="dataset",
    )


def _hf_get_snapshot_dir(repo_id, cache_dir=HF_DATASETS_DIR):
    repo_cache_dir = Path(cache_dir) / f"datasets--{repo_id.replace('/', '--')}"
    snapshots_dir = repo_cache_dir / "snapshots"
    if not snapshots_dir.exists() or not any(snapshots_dir.iterdir()):
        # Try to trigger a download to populate the cache
        files = _hf_list_files(repo_id)
        # Pick the first file (prefer .h5/.hdf5 if possible)
        h5_files = [f for f in files if f.endswith(".h5") or f.endswith(".hdf5")]
        target_file = h5_files[0] if h5_files else files[0]
        _hf_download(repo_id, target_file, cache_dir)
        # Now try again
        if not snapshots_dir.exists() or not any(snapshots_dir.iterdir()):
            raise FileNotFoundError(
                f"No snapshots found in Hugging Face cache for {repo_id} after download attempt"
            )
    snapshot_hashes = sorted(snapshots_dir.iterdir(), key=lambda p: p.stat().st_mtime, reverse=True)
    if not snapshot_hashes:
        raise FileNotFoundError(f"No snapshot found for {repo_id} in cache.")
    return snapshot_hashes[0]


def _hf_resolve_path(hf_path: str, cache_dir=HF_DATASETS_DIR):
    """Download a file or directory from Hugging Face Hub to a local cache directory.
    Returns the local path to the downloaded file or directory.
    """
    repo_id, subpath = _hf_parse_path(hf_path)
    files = _hf_list_files(repo_id)
    snapshot_dir = _hf_get_snapshot_dir(repo_id, cache_dir)

    def is_h5(f):
        return f.endswith(".h5") or f.endswith(".hdf5")

    if subpath:
        # Directory
        if any(f.startswith(subpath + "/") for f in files):
            local_dir = snapshot_dir / subpath
            for f in files:
                if f.startswith(subpath + "/") and is_h5(f):
                    _hf_download(repo_id, f, cache_dir)
            if not local_dir.exists():
                raise FileNotFoundError(f"Directory {local_dir} not found after download.")
            return local_dir
        # File
        elif any(f == subpath for f in files) and is_h5(subpath):
            _hf_download(repo_id, subpath, cache_dir)
            local_file = snapshot_dir / subpath
            if not local_file.exists():
                raise FileNotFoundError(f"File {local_file} not found after download.")
            return local_file
        else:
            raise FileNotFoundError(f"{subpath} not found in {repo_id}")
    else:
        # All .h5/.hdf5 files in repo
        for f in files:
            if is_h5(f):
                _hf_download(repo_id, f, cache_dir)
        return snapshot_dir



"""``zea``: *A Toolbox for Cognitive Ultrasound Imaging.*"""

import importlib.util
import os

from . import log

# dynamically add __version__ attribute (see pyproject.toml)
# __version__ = __import__("importlib.metadata").metadata.version(__package__)
__version__ = "0.0.1"


def setup():
    """Setup function to initialize the zea package."""

    def _check_backend_installed():
        """Assert that at least one ML backend (torch, tensorflow, jax) is installed.
        If not, raise an AssertionError with a helpful install message.
        """

        ml_backends = ["torch", "tensorflow", "jax"]
        for backend in ml_backends:
            if importlib.util.find_spec(backend) is not None:
                return

        backend_env = os.environ.get("KERAS_BACKEND", "numpy")
        install_guide_urls = {
            "torch": "https://pytorch.org/get-started/locally/",
            "tensorflow": "https://www.tensorflow.org/install",
            "jax": "https://docs.jax.dev/en/latest/installation.html",
        }
        guide_url = install_guide_urls.get(backend_env, "https://keras.io/getting_started/")
        raise ImportError(
            "No ML backend (torch, tensorflow, jax) installed in current environment. "
            f"Please install at least one ML backend before importing {__package__} or "
            f"any other library. Current KERAS_BACKEND is set to '{backend_env}', "
            f"please install it first, see: {guide_url}. One simple alternative is to "
            f"install with default backend: `pip install {__package__}[jax]`."
        )

    _check_backend_installed()

    import keras

    log.info(f"Using backend {keras.backend.backend()!r}")


# call and clean up namespace
setup()
del setup

from . import (
    agent,
    beamform,
    data,
    display,
    io_lib,
    metrics,
    models,
    simulator,
    tensor_ops,
    utils,
    visualize,
)
from .config import Config
from .data.datasets import Dataset, Folder
from .data.file import File, load_file
from .datapaths import set_data_paths
from .interface import Interface
from .internal.device import init_device
from .internal.setup_zea import set_backend, setup, setup_config
from .ops import Pipeline
from .probes import Probe
from .scan import Scan



"""
Mask generation utilities.

These masks are used as a measurement operator for focused scan-line subsampling.
"""

from typing import List

import keras
from keras import ops

from zea import tensor_ops
from zea.agent.gumbel import hard_straight_through

_DEFAULT_DTYPE = "bool"


def indices_to_k_hot(
    indices: List[int],
    n_possible_actions: int,
    dtype=_DEFAULT_DTYPE,
):
    """Convert a list of indices to a k-hot encoded vector.

    A k-hot encoded vector is suitable during tracing when the number of actions can change.
    This is the default represenation for actions in zea.

    Args:
        indices (List[int]): List of indices to set to 1.
        n_possible_actions (int): Total number of possible actions.
        dtype (str, optional): Data type of the mask. Defaults to _DEFAULT_DTYPE.

    Returns:
        Tensor: k-hot-encoded vector of shape (n_possible_actions).
    """
    mask = ops.zeros(n_possible_actions, dtype=dtype)
    return ops.scatter_update(
        mask, ops.expand_dims(indices, axis=1), ops.ones(len(indices), dtype=dtype)
    )


def k_hot_to_indices(selected_lines, n_actions: int, fill_value=-1):
    """Convert k-hot encoded lines to indices of selected actions.

    Args:
        selected_lines (Tensor): k-hot encoded lines of shape (batch_size, n_possible_actions).
        n_actions (int): Number of lines selected.
        fill_value (int, optional): Value to fill in case there are not enough selected actions.
            Defaults to -1.

    Returns:
        Tensor: Indices of selected actions of shape (batch_size, n_actions).
            If there are fewer than `n_actions` selected, the remaining indices will be
            filled with `fill_value`.
    """

    # Find nonzero indices for each frame
    def get_nonzero(row):
        return tensor_ops.nonzero(row > 0, size=n_actions, fill_value=fill_value)[0]

    indices = ops.vectorized_map(get_nonzero, selected_lines)
    return indices


def random_uniform_lines(
    n_actions: int,
    n_possible_actions: int,
    n_masks: int,
    seed: int | keras.random.SeedGenerator = None,
    dtype=_DEFAULT_DTYPE,
):
    """Will generate a mask with random lines.

    Guarantees precisely n_actions.

    Args:
        n_actions (int): Number of actions to be selected.
        n_possible_actions (int): Number of possible actions.
        n_masks (int): Number of masks to generate.
        seed (int | SeedGenerator | jax.random.key, optional): Seed for random number generation.
            Defaults to None.

    Returns:
        Tensor: k-hot-encoded line vectors of shape (n_masks, n_possible_actions).
                Needs to be converted to image size.
    """
    masks = keras.random.uniform([n_masks, n_possible_actions], seed=seed, dtype="float32")
    masks = hard_straight_through(masks, n_actions)
    return ops.cast(masks, dtype=dtype)


def _assert_equal_spacing(n_actions, n_possible_actions):
    assert n_possible_actions % n_actions == 0, (
        "Number of actions must divide evenly into possible actions to use equispaced sampling."
    )


def initial_equispaced_lines(
    n_actions, n_possible_actions, dtype=_DEFAULT_DTYPE, assert_equal_spacing=True
):
    """Generate an initial equispaced k-hot line mask.

    For example, if ``n_actions=2`` and ``n_possible_actions=6``,
    then ``initial_mask=[1, 0, 0, 1, 0, 0]``.

    Args:
        n_actions (int): Number of actions to be selected.
        n_possible_actions (int): Number of possible actions.
        dtype (str, optional): Data type of the mask. Defaults to _DEFAULT_DTYPE.
        assert_equal_spacing (bool, optional): If True, asserts that
            `n_possible_actions` is divisible by `n_actions`, this means that every
            line will have the exact same spacing. Otherwise, there might be
            some spacing differences. Defaults to True.

    Returns:
        Tensor: k-hot-encoded line vector of shape (n_possible_actions).
            Needs to be converted to image size.
    """
    if assert_equal_spacing:
        _assert_equal_spacing(n_actions, n_possible_actions)
        selected_indices = ops.arange(0, n_possible_actions, n_possible_actions // n_actions)
    else:
        selected_indices = ops.linspace(0, n_possible_actions - 1, n_actions, dtype="int32")

    return indices_to_k_hot(selected_indices, n_possible_actions, dtype=dtype)


def next_equispaced_lines(previous_lines, shift=1):
    """
    Rolls the previous equispaced mask of shape (..., n_possible_actions) to the right by
    `shift` which is 1 by default.
    """
    return ops.roll(previous_lines, shift=shift, axis=-1)


def lines_to_im_size(lines, img_size: tuple):
    """
    Convert k-hot-encoded line vectors to image size.

    Args:
        lines (Tensor): shape is (n_masks, n_possible_actions)
        img_size (tuple): (height, width)

    Returns:
        Tensor: Masks of shape (n_masks, img_size, img_size)
    """
    height, width = img_size

    remainder = width % ops.shape(lines)[1]
    assert remainder == 0, (
        f"Width must be divisible by number of actions. Got remainder: {remainder}."
    )

    # Repeat till width of image
    masks = ops.repeat(lines, width // ops.shape(lines)[1], axis=1)

    # Repeat till height of image
    masks = ops.repeat(masks[:, None], height, axis=1)

    return masks


def make_line_mask(
    line_indices: List[int],
    image_shape: List[int],
    line_width: int = 1,
    dtype=_DEFAULT_DTYPE,
):
    """
    Creates a mask with vertical (i.e. second axis) lines at specified indices.

    Args:
        line_indices (List[int]): A list of indices where the lines should be drawn.
        image_shape (List[int]): The shape of the image as [height, width, channels].
        line_width (int, optional): The width of each line. Defaults to 1.
        dtype (str, optional): The data type of the mask. Defaults to "float32".

    Returns:
        mask (Tensor): A tensor of the same shape as `image_shape` with lines drawn
            at the specified indices.
    """
    height, width, channels = image_shape

    # Create k-hot vector for the line indices
    k_hot = indices_to_k_hot(line_indices, width // line_width, dtype=dtype)
    # Expand to (1, n_possible_actions) for lines_to_im_size
    k_hot = ops.expand_dims(k_hot, axis=0)
    # Use lines_to_im_size to create the mask of shape (1, height, width)
    mask_2d = lines_to_im_size(k_hot, (height, width))[0]

    # Expand to (height, width, channels)
    return ops.broadcast_to(mask_2d[..., None], (height, width, channels))



"""zea H5 file functionality."""

import enum
from pathlib import Path
from typing import List

import h5py
import numpy as np

from zea import log
from zea.data.preset_utils import HF_PREFIX, _hf_resolve_path
from zea.internal.checks import (
    _DATA_TYPES,
    _NON_IMAGE_DATA_TYPES,
    _REQUIRED_SCAN_KEYS,
    get_check,
)
from zea.probes import Probe
from zea.scan import Scan
from zea.utils import reduce_to_signature


def assert_key(file: h5py.File, key: str):
    """Asserts key is in a h5py.File."""
    if key not in file.keys():
        raise KeyError(f"{key} not found in file")


class File(h5py.File):
    """h5py.File in zea format."""

    def __init__(self, name, *args, **kwargs):
        """Initialize the file.

        Args:
            name (str, Path, HFPath): The path to the file.
                Can be a string or a Path object. Additionally can be a string with
                the prefix 'hf://', in which case it will be resolved to a
                huggingface path.
            *args: Additional arguments to pass to h5py.File.
            **kwargs: Additional keyword arguments to pass to h5py.File.
        """

        if str(name).startswith(HF_PREFIX):
            name = _hf_resolve_path(str(name))

        if "locking" not in kwargs and "mode" in kwargs and kwargs["mode"] == "r":
            # If the file is opened in read mode, disable locking
            kwargs["locking"] = False

        super().__init__(name, *args, **kwargs)

    @property
    def path(self):
        """Return the path of the file."""
        return Path(self.filename)

    @property
    def name(self):
        """Return the name of the file."""
        return self.path.name

    @property
    def stem(self):
        """Return the stem of the file."""
        return self.path.stem

    @property
    def event_keys(self):
        """Return all events in the file."""
        return [key for key in self.keys() if "event" in key]

    @property
    def has_events(self):
        """Check if the file has events."""
        return any("event" in key for key in self.keys())
        # return self.attrs.get("event_structure", False)

    @property
    def n_frames(self):
        """Return number of frames in a file."""

        if "scan" in self.file:
            return int(self.file["scan"]["n_frames"][()])
        else:
            return sum(int(event["scan"]["n_frames"][()]) for event in self.file.values())

    def get_event_shapes(self, key):
        """Get the shapes of a key for all events."""
        for event_key in self.event_keys:
            yield self[event_key][key].shape

    def events_have_same_shape(self, key):
        """Check if all events have the same shape for a given key."""
        if not self.has_events:
            return True

        shapes = list(self.get_event_shapes(key))
        return len(np.unique(shapes)) == 1

    def _simple_index(self, key):
        return not self.has_events or "event" in key

    def shape(self, key) -> tuple:
        """Return shape of some key, or all events."""
        key = self.format_key(key)

        if self._simple_index(key):
            return self[key].shape
        else:
            raise NotImplementedError

    @staticmethod
    def _prepare_indices(indices):
        """Prepare the indices for loading data from hdf5 files.
        Options:
            - str("all")
            - int -> single frame
            - list of ints -> indexes first axis (frames)
            - list of list, ranges or slices -> indexes multiple axes

        Returns:
            indices (tuple): A tuple of indices / slices to use for indexing.
        """
        _value_error_msg = (
            f"Invalid value for indices: {indices}. "
            "Indices can be a 'all', int or a List[int, tuple, list, slice, range]."
        )

        # Check all options that only index the first axis
        if isinstance(indices, str):
            if indices == "all":
                return slice(None)
            else:
                raise ValueError(_value_error_msg)

        if isinstance(indices, range):
            return list(indices)

        if isinstance(indices, (int, slice, np.integer)):
            return indices

        # At this point, indices should be a list or tuple
        assert isinstance(indices, (list, tuple, np.ndarray)), _value_error_msg

        assert all(
            isinstance(idx, (list, tuple, int, slice, range, np.ndarray, np.integer))
            for idx in indices
        ), _value_error_msg

        # Convert ranges to lists
        processed_indices = [list(idx) if isinstance(idx, range) else idx for idx in indices]

        # Check if items are list-like and cast to tuple (needed for hdf5)
        if any(isinstance(idx, (list, tuple, slice)) for idx in processed_indices):
            processed_indices = tuple(processed_indices)

        return processed_indices

    def load_scan(self, event=None):
        """Alias for get_scan_parameters."""
        return self.get_scan_parameters(event)

    @staticmethod
    def check_data(data, key):
        """Check the data for a given key. For example, will check if the shape matches
        the data type (such as raw_data, ...)"""
        if key in _DATA_TYPES:
            get_check(key)(data, with_batch_dim=None)

    def format_key(self, key):
        """Format the key to match the data type."""
        # TODO: support events

        if isinstance(key, enum.Enum):
            key = key.value

        # Return the key if it is in the file
        if key in self.keys():
            return key

        # Add 'data/' prefix if not present
        if "data/" not in key:
            key = "data/" + key

        assert key in self.keys(), (
            f"Key {key} not found in file. Available keys: {list(self['data'].keys())}"
        )

        return key

    def to_iterator(self, key):
        """Convert the data to an iterator over all frames."""
        for frame_idx in range(self.n_frames):
            yield self.load_data(key, frame_idx)

    @staticmethod
    def key_to_data_type(key):
        """Convert the key to a data type."""
        data_type = key.split("/")[-1]
        return data_type

    def load_transmits(self, key, selected_transmits):
        """Load raw_data or aligned_data for a given list of transmits.
        Args:
            data_type (str): The type of data to load. Options are 'raw_data' and 'aligned_data'.
            selected_transmits (list, np.ndarray): The transmits to load.
        """
        key = self.format_key(key)
        data_type = self.key_to_data_type(key)
        assert data_type in ["raw_data", "aligned_data"], (
            f"Cannot load transmits for {data_type}. Only raw_data and aligned_data are supported."
        )
        indices = [slice(None), np.array(selected_transmits)]
        return self.load_data(key, indices)

    def load_data(self, data_type, indices: str | int | List[int] = "all"):
        """Load data from the file.

        Args:
            data_type (str): The type of data to load. Options are 'raw_data', 'aligned_data',
                'beamformed_data', 'envelope_data', 'image' and 'image_sc'.
            indices (str, int, list, optional): The indices to load. Defaults to "all" in
                which case all frames are loaded. If an int is provided, it will be used
                as a single index. If a list is provided, it will be used as a list of
                indices.
        """
        key = self.format_key(data_type)
        indices = self._prepare_indices(indices)

        if self._simple_index(key):
            data = self[key]
            try:
                data = data[indices]
            except (OSError, IndexError) as exc:
                raise ValueError(
                    f"Invalid indices {indices} for key {key}. {key} has shape {data.shape}."
                ) from exc
            self.check_data(data, key)
        elif self.events_have_same_shape(key):
            raise NotImplementedError
        else:
            raise NotImplementedError

        return data

    @property
    def probe_name(self):
        """Reads the probe name from the data file and returns it."""
        assert "probe" in self.attrs, (
            "Probe name not found in file attributes. "
            "Make sure you are using a zea file. "
            f"Found attributes: {list(self.attrs)}"
        )
        probe_name = self.attrs["probe"]
        return probe_name

    @property
    def description(self):
        """Reads the description from the data file and returns it."""
        assert "description" in self.attrs, (
            "Description not found in file attributes. "
            "Make sure you are using a zea file. "
            f"Found attributes: {list(self.attrs)}"
        )
        description = self.attrs["description"]
        return description

    def get_parameters(self, event=None):
        """Returns a dictionary of parameters to initialize a scan
        object that comes with the file (stored inside datafile).

        If there are no scan parameters in the hdf5 file, returns
        an empty dictionary.

        Args:
            event (int, optional): Event number. When specified, an event structure
                is expected as follows::

                    event_0 / scan
                    event_1 / scan
                    ...

                Defaults to None. In that case no event structure is expected.

        Returns:
            dict: The scan parameters.
        """
        scan_parameters = {}
        if "scan" in self:
            scan_parameters = recursively_load_dict_contents_from_group(self, "scan")
        elif "event" in list(self.keys())[0]:
            if event is None:
                raise ValueError(
                    log.error(
                        "Please specify an event number to read scan parameters "
                        "from a file with an event structure."
                    )
                )

            assert f"event_{event}/scan" in self, (
                f"Could not find scan parameters for event {event} in file. "
                f"Found number of events: {len(self.keys())}."
            )

            scan_parameters = recursively_load_dict_contents_from_group(self, f"event_{event}/scan")
        else:
            log.warning("Could not find scan parameters in file.")

        return scan_parameters

    def get_scan_parameters(self, event=None) -> dict:
        """Returns a dictionary of default parameters to initialize a scan
        object that works with the file.

        Returns:
            dict: The default parameters (the keys are identical to the
                __init__ parameters of the Scan class).
        """
        file_scan_parameters = self.get_parameters(event)

        scan_parameters = {}
        for parameter, value in file_scan_parameters.items():
            if parameter in Scan.VALID_PARAMS:
                param_type = Scan.VALID_PARAMS[parameter]["type"]
                if param_type in (bool, int, float):
                    scan_parameters[parameter] = param_type(value)
                elif isinstance(param_type, tuple) and float in param_type:
                    scan_parameters[parameter] = float(value)
                else:
                    scan_parameters[parameter] = value

        if len(scan_parameters) == 0:
            log.info(f"Could not find proper scan parameters in {self}.")
        return scan_parameters

    def scan(self, event=None, **kwargs) -> Scan:
        """Returns a Scan object initialized with the parameters from the file.

        Args:
            event (int, optional): Event number. When specified, an event structure
                is expected as follows::

                    event_0 / scan
                    event_1 / scan
                    ...

                Defaults to None. In that case no event structure is expected.
            **kwargs: Additional keyword arguments to pass to the Scan object.
                These will override the parameters from the file if they are
                present in the file.

        Returns:
            Scan: The scan object.
        """
        return Scan.merge(self.get_scan_parameters(event), kwargs)

    def get_probe_parameters(self, event=None) -> dict:
        """Returns a dictionary of probe parameters to initialize a probe
        object that comes with the file (stored inside datafile).

        Returns:
            dict: The probe parameters.
        """
        file_scan_parameters = self.get_parameters(event)

        probe_parameters = reduce_to_signature(
            Probe.from_name("generic").__init__, file_scan_parameters
        )
        return probe_parameters

    def probe(self, event=None) -> Probe:
        """Returns a Probe object initialized with the parameters from the file.

        Args:
            event (int, optional): Event number. When specified, an event structure
                is expected as follows::

                    event_0 / scan
                    event_1 / scan
                    ...

                Defaults to None. In that case, no event structure is expected.

        Returns:
            Probe: The probe object.
        """
        probe_parameters = self.get_probe_parameters(event)
        if self.probe_name == "generic":
            return Probe.from_name(self.probe_name, **probe_parameters)
        else:
            probe = Probe.from_name(self.probe_name)

            probe_geometry = probe_parameters.get("probe_geometry", None)
            if not np.allclose(probe_geometry, probe.probe_geometry):
                probe.probe_geometry = probe_geometry
                log.warning(
                    "The probe geometry in the data file does not "
                    "match the probe geometry of the probe. The probe "
                    "geometry has been updated to match the data file."
                )
        return probe

    def recursively_load_dict_contents_from_group(self, path: str, squeeze: bool = False) -> dict:
        """Load dict from contents of group

        Values inside the group are converted to numpy arrays
        or primitive types (int, float, str). Single element
        arrays are converted to the corresponding primitive type (if squeeze=True)

        Args:
            path (str): path to group
            squeeze (bool, optional): squeeze arrays with single element.
                Defaults to False.
        Returns:
            dict: dictionary with contents of group
        """
        return recursively_load_dict_contents_from_group(self, path, squeeze)

    @classmethod
    def get_shape(cls, path: str, key: str) -> tuple:
        """Get the shape of a key in a file.

        Args:
            path (str): The path to the file.
            key (str): The key to get the shape of.

        Returns:
            tuple: The shape of the key.
        """
        with cls(path, mode="r") as file:
            return file.shape(key)

    def validate(self):
        """Validate the file structure.

        Returns:
            dict: A dictionary with the validation results.
        """
        return validate_file(file=self)

    def __repr__(self):
        return (
            f"<zea.data.file.File at 0x{id(self):x} "
            f'("{Path(self.filename).name}" mode={self.mode})>'
        )

    def __str__(self):
        return f"zea HDF5 File: '{self.path.name}' (mode={self.mode})"

    def summary(self):
        """Print the contents of the file."""
        _print_hdf5_attrs(self)


def load_file(
    path,
    data_type="raw_data",
    indices: str | int | List[int] = "all",
    scan_kwargs: dict = None,
):
    """Loads a zea data files (h5py file).

    Returns the data together with a scan object containing the parameters
    of the acquisition and a probe object containing the parameters of the probe.

    Additionally, it can load a specific subset of frames / transmits.

    # TODO: add support for event

    Args:
        path (str, pathlike): The path to the hdf5 file.
        data_type (str, optional): The type of data to load. Defaults to
            'raw_data'. Other options are 'aligned_data', 'beamformed_data',
            'envelope_data', 'image' and 'image_sc'.
        indices (str, int, list, optional): The indices to load. Defaults to "all" in
            which case all frames are loaded. If an int is provided, it will be used
            as a single index. If a list is provided, it will be used as a list of
            indices.
        scan_kwargs (Config, dict, optional): Additional keyword arguments
            to pass to the Scan object. These will override the parameters from the file
            if they are present in the file. Defaults to None.

    Returns:
        (np.ndarray): The raw data of shape (n_frames, n_tx, n_ax, n_el, n_ch).
        (Scan): A scan object containing the parameters of the acquisition.
        (Probe): A probe object containing the parameters of the probe.
    """
    # Define the additional keyword parameters from the scan object
    if scan_kwargs is None:
        scan_kwargs = {}

    with File(path, mode="r") as file:
        # Load the probe object from the file
        probe = file.probe()

        # Load the desired frames from the file
        data = file.load_data(data_type, indices=indices)

        # extract transmits from indices
        # we only have to do this when the data has a n_tx dimension
        # in that case we also have update scan parameters to match
        # the number of selected transmits
        if data_type in ["raw_data", "aligned_data"]:
            indices = File._prepare_indices(indices)
            n_tx = data.shape[1]
            if isinstance(indices, tuple) and len(indices) > 1:
                tx_idx = indices[1]
                transmits = np.arange(n_tx)[tx_idx]
                scan_kwargs["selected_transmits"] = transmits

        scan = file.scan(**scan_kwargs)

        return data, scan, probe


def recursively_load_dict_contents_from_group(
    h5file: h5py._hl.files.File, path: str, squeeze: bool = False
) -> dict:
    """Load dict from contents of group

    Values inside the group are converted to numpy arrays
    or primitive types (int, float, str). Single element
    arrays are converted to the corresponding primitive type (if squeeze=True)

    Args:
        h5file (h5py._hl.files.File): h5py file object
        path (str): path to group
        squeeze (bool, optional): squeeze arrays with single element.
            Defaults to False.
    Returns:
        dict: dictionary with contents of group
    """
    ans = {}
    for key, item in h5file[path].items():
        if isinstance(item, h5py._hl.dataset.Dataset):
            ans[key] = item[()]
            # all ones in shape
            if squeeze:
                if ans[key].shape == () or all(i == 1 for i in ans[key].shape):
                    # check for strings
                    if isinstance(ans[key], str):
                        ans[key] = str(ans[key])
                    # check for integers
                    elif int(ans[key]) == float(ans[key]):
                        ans[key] = int(ans[key])
                    else:
                        ans[key] = float(ans[key])
        elif isinstance(item, h5py._hl.group.Group):
            ans[key] = recursively_load_dict_contents_from_group(h5file, path + "/" + key + "/")
    return ans


def _print_hdf5_attrs(hdf5_obj, prefix=""):
    """Recursively prints all keys, attributes, and shapes in an HDF5 file.

    Args:
        hdf5_obj (h5py.File, h5py.Group, h5py.Dataset): HDF5 object to print.
        prefix (str, optional): Prefix to print before each line. This
            parameter is used in internal recursion and should not be supplied
            by the user.
    """
    assert isinstance(hdf5_obj, (h5py.File, h5py.Group, h5py.Dataset)), (
        "ERROR: hdf5_obj must be a File, Group, or Dataset object"
    )

    if isinstance(hdf5_obj, h5py.File):
        name = "root" if hdf5_obj.name == "/" else hdf5_obj.name
        print(prefix + name + "/")
        prefix += "    "
    elif isinstance(hdf5_obj, h5py.Dataset):
        shape_str = str(hdf5_obj.shape).replace(",)", ")")
        print(prefix + "├── " + hdf5_obj.name + " (shape=" + shape_str + ")")
        prefix += "│   "

    # Print all attributes
    for key, val in hdf5_obj.attrs.items():
        print(prefix + "├── " + key + ": " + str(val))

    # Recursively print all keys, attributes, and shapes in groups
    if isinstance(hdf5_obj, h5py.Group):
        for i, key in enumerate(hdf5_obj.keys()):
            is_last = i == len(hdf5_obj.keys()) - 1
            if is_last:
                marker = "└── "
                new_prefix = prefix + "    "
            else:
                marker = "├── "
                new_prefix = prefix + "│   "
            print(prefix + marker + key + "/")
            _print_hdf5_attrs(hdf5_obj[key], new_prefix)


def validate_file(path: str = None, file: File = None):
    """Reads the hdf5 file at the given path and validates its structure.

    Provide either the path or the file, but not both.

    Args:
        path (str, pathlike): The path to the hdf5 file.
        file (File): The hdf5 file.

    """
    assert (path is not None) ^ (file is not None), (
        "Provide either the path or the file, but not both."
    )

    if path is not None:
        path = Path(path)
        with File(path, "r") as _file:
            event_structure, num_events = _validate_hdf5_file(_file)
    else:
        event_structure, num_events = _validate_hdf5_file(file)

    return {
        "status": "success",
        "event_structure": event_structure,
        "num_events": num_events,
    }


def _validate_hdf5_file(file: File):
    all_keys = list(file.keys())

    if file.has_events:
        num_events = len(all_keys)
        for event_no in range(num_events):
            assert_key(file, f"event_{event_no}")
            _validate_structure(file[f"event_{event_no}"])
    else:
        num_events = 0
        _validate_structure(file)

    return file.has_events, num_events


def _validate_structure(file: File):
    # Validate the root group
    assert_key(file, "data")

    # Assert file["data"] is a group
    assert isinstance(file["data"], h5py.Group), (
        "The data group is not a group. Please check the file structure. "
        "Maybe this is not a zea file?"
    )

    # Check if there is only image data
    not_only_image_data = len([i for i in _NON_IMAGE_DATA_TYPES if i in file["data"].keys()]) > 0

    # Only check scan group if there is non-image data
    if not_only_image_data:
        assert_key(file, "scan")

        for key in _REQUIRED_SCAN_KEYS:
            assert_key(file["scan"], key)

    # validate the data group
    for key in file["data"].keys():
        assert key in _DATA_TYPES, "The data group contains an unexpected key."

        # Validate data shape
        data_shape = file["data"][key].shape
        if key == "raw_data":
            get_check(key)(shape=data_shape, with_batch_dim=True)
            assert data_shape[0] == file["scan"]["n_frames"][()], (
                "n_frames does not match the first dimension of raw_data."
            )
            assert data_shape[1] == file["scan"]["n_tx"][()], (
                "n_tx does not match the second dimension of raw_data."
            )
            assert data_shape[2] == file["scan"]["n_ax"][()], (
                "n_ax does not match the third dimension of raw_data."
            )
            assert data_shape[3] == file["scan"]["n_el"][()], (
                "n_el does not match the fourth dimension of raw_data."
            )
        elif key == "aligned_data":
            get_check(key)(shape=data_shape, with_batch_dim=True)
            assert data_shape[0] == file["scan"]["n_frames"][()], (
                "n_frames does not match the first dimension of aligned_data."
            )
        elif key == "beamformed_data":
            get_check(key)(shape=data_shape, with_batch_dim=True)
            assert data_shape[0] == file["scan"]["n_frames"][()], (
                "n_frames does not match the first dimension of beamformed_data."
            )
        elif key == "envelope_data":
            get_check(key)(shape=data_shape, with_batch_dim=True)
            assert data_shape[0] == file["scan"]["n_frames"][()], (
                "n_frames does not match the first dimension of envelope_data."
            )
        elif key == "image":
            get_check(key)(shape=data_shape, with_batch_dim=True)
            assert data_shape[0] == file["scan"]["n_frames"][()], (
                "n_frames does not match the first dimension of image."
            )
        elif key == "image_sc":
            get_check(key)(shape=data_shape, with_batch_dim=True)
            assert data_shape[0] == file["scan"]["n_frames"][()], (
                "n_frames does not match the first dimension of image_sc."
            )

    if not_only_image_data:
        _assert_scan_keys_present(file)

    _assert_unit_and_description_present(file)


def _assert_scan_keys_present(file: File):
    """Ensure that all required keys are present.

    Args:
        file (h5py.File): The file instance to check.

    Raises:
        AssertionError: If a required key is missing or does not have the right shape.
    """
    for required_key in _REQUIRED_SCAN_KEYS:
        assert required_key in file["scan"].keys(), (
            f"The scan group does not contain the required key {required_key}."
        )

    # Ensure that all keys have the correct shape
    for key in file["scan"].keys():
        if isinstance(file["scan"][key], h5py.Group):
            shape_file = None
        else:
            shape_file = file["scan"][key].shape

        if key == "probe_geometry":
            correct_shape = (file["scan"]["n_el"][()], 3)

        elif key == "t0_delays":
            correct_shape = (
                file["scan"]["n_tx"][()],
                file["scan"]["n_el"][()],
            )
        elif key == "tx_apodizations":
            correct_shape = (
                file["scan"]["n_tx"][()],
                file["scan"]["n_el"][()],
            )

        elif key == "focus_distances":
            correct_shape = (file["scan"]["n_tx"][()],)

        elif key == "polar_angles":
            correct_shape = (file["scan"]["n_tx"][()],)

        elif key == "azimuth_angles":
            correct_shape = (file["scan"]["n_tx"][()],)

        elif key == "initial_times":
            correct_shape = (file["scan"]["n_tx"][()],)

        elif key == "time_to_next_transmit":
            correct_shape = (
                file["scan"]["n_frames"][()],
                file["scan"]["n_tx"][()],
            )
        elif key == "tgc_gain_curve":
            correct_shape = (file["scan"]["n_ax"][()],)
        elif key == "tx_waveform_indices":
            correct_shape = (file["scan"]["n_tx"][()],)
        elif key in ("waveforms_one_way", "waveforms_two_way"):
            correct_shape = None

        elif key in (
            "sampling_frequency",
            "center_frequency",
            "n_frames",
            "n_tx",
            "n_el",
            "n_ax",
            "n_ch",
            "sound_speed",
            "bandwidth_percent",
            "element_width",
            "lens_correction",
        ):
            correct_shape = ()
            shape_file = file["scan"][key].shape

        else:
            correct_shape = None
            log.warning(f"No validation has been defined for {log.orange(key)}.")

        if correct_shape is not None:
            assert shape_file == correct_shape, (
                f"`{key}` does not have the correct shape. "
                f"Expected shape: {correct_shape}, got shape: {shape_file}"
            )


def _assert_unit_and_description_present(hdf5_file, _prefix=""):
    """Checks that all keys have a unit and description attribute.

    Args:
        hdf5_file (h5py.File): The hdf5 file to check.

    Raises:
        AssertionError: If a file does not have a unit or description attribute.
    """
    for key in hdf5_file.keys():
        if isinstance(hdf5_file[key], h5py.Group):
            _assert_unit_and_description_present(hdf5_file[key], _prefix=_prefix + key + "/")
        else:
            assert "unit" in hdf5_file[key].attrs.keys(), (
                f"The file {_prefix}/{key} does not have a unit attribute."
            )
            assert "description" in hdf5_file[key].attrs.keys(), (
                f"The file {_prefix}/{key} does not have a description attribute."
            )



"""Wrapper module to load tf2jax if available."""

try:
    from tf2jax import *  # noqa: F401, F403
except ImportError as exc:
    raise ImportError(
        "tf2jax is not installed. "
        "Suggested installation: `pip install tf2jax==0.3.6 && pip install keras -U`"
        "Note that this may install a newer version of Keras than the one you have installed! "
    ) from exc



"""Generative models for zea."""

import abc

from zea.models.base import BaseModel


class GenerativeModel(abc.ABC):
    """Abstract base class for generative models."""

    def fit(self, data, **kwargs):
        """Fit the model to the data.

        Args:
            data: The data to fit the model to.
            **kwargs: Additional arguments to pass to the fitting procedure.
        """
        raise NotImplementedError("fit() must be implemented in subclasses.")

    def sample(self, n_samples=1, **kwargs):
        r"""Draw samples $x \sim p(x)$ from the model.

        Args:
            n_samples: Number of samples to generate.
            **kwargs: Additional arguments to pass to the sampling procedure.

        Returns:
            Samples $x$ from the model distribution $p(x)$.
        """
        raise NotImplementedError("sample() must be implemented in subclasses.")

    def posterior_sample(self, measurements, n_samples=1, **kwargs):
        r"""Draw samples $z \sim p(z \mid x)$ from the posterior given measurements.

        Args:
            measurements: The measurements $x$ to condition the posterior on.
            n_samples: Number of posterior samples to generate. This will add
                an additional dimension to the output. For instance,
                if `measurements` has shape `(batch_size, ...)`, the output will
                have shape `(batch_size, n_samples, ...)`.
            **kwargs: Additional arguments to pass to the sampling procedure.

        Returns:
            Samples $z$ from the posterior $p(z \mid x)$.
        """
        raise NotImplementedError("posterior_sample() must be implemented in subclasses.")

    def log_density(self, data, **kwargs):
        r"""Compute the log-density $\log p(x)$ of the data under the model.

        Args:
            data: The data $x$ to compute the log-density for.
            **kwargs: Additional arguments.

        Returns:
            Log-density $\log p(x)$ of the data.
        """
        raise NotImplementedError("log_density() must be implemented in subclasses.")


class DeepGenerativeModel(BaseModel, GenerativeModel):
    """Base class for deep generative models.

    Inherits from both GenerativeModel and BaseModel to combine
    generative capabilities with Keras model functionality.
    """

    def __init__(self, name="deep_generative_model", **kwargs):
        """Initialize a deep generative model.

        Args:
            name: Name of the model.
            **kwargs: Additional arguments to pass to BaseModel.
        """
        BaseModel.__init__(self, name=name, **kwargs)



"""Functionality to convert Verasonics matlab raw files to the zea format.

Example (MATLAB):

    .. code-block:: matlab

        >> setup_script;
        >> VSX;
        >> save_raw('C:/path/to/raw_data.mat');

Then in python:

    .. code-block:: python

        from zea.data_format.zea_from_matlab_raw import zea_from_matlab_raw

        zea_from_matlab_raw("C:/path/to/raw_data.mat", "C:/path/to/output.hdf5")

Or alternatively, use the script below to convert all .mat files in a directory:

    .. code-block:: bash

        python zea/data/convert/matlab.py "C:/path/to/directory"

or without the directory argument, the script will prompt you to select a directory
using a file dialog.

Event structure
---------------

By default the zea dataformat saves all the data to an hdf5 file with the following structure:

.. code-block:: text

    regular_zea_dataset.hdf5
    ├── data
    └── scan
          └── center_frequency: 1MHz

The data is stored in the ``data`` group and the scan parameters are stored in the ``scan``.
However, when we do an adaptive acquisition, some scanning parameters might change. These
blocks of data with consistent scanning parameters we call events. In the case we have multiple
events, we store the data in the following structure:

.. code-block:: text

    zea_dataset.hdf5
    ├── event_0
    │   ├── data
    │   └── scan
    │       └── center_frequency: 1MHz
    ├── event_1
    │   ├── data
    │   └── scan
    │       └── center_frequency: 2MHz
    ├── event_2
    │   ├── data
    │   └── scan
    └── event_3
        ├── data
        └── scan

This structure is supported by the zea toolbox. The way we can save the data in this structure
from the Verasonics, is by changing the setup script to keep track of the TX struct at each event.

The way this is done is still in development, an example of such an acquisition script that is
compatible with saving event structures is found here:
`setup_agent.m <https://github.com/tue-bmd/needle-tracking/blob/ius2024-demo-nc/verasonics/setup_agent.m>`_

Adding additional elements
--------------------------

You can add additional elements to the dataset by defining a function that reads the
data from the file and returns a ``DatasetElement``. Then pass the function to the
``zea_from_matlab_raw`` function as a list.

.. code-block:: python

    def read_max_high_voltage(file):
        lens_correction = file["Trans"]["lensCorrection"][0, 0].item()
        return lens_correction


    def read_high_voltage_func(file):
        return DatasetElement(
            group_name="scan",
            dataset_name="max_high_voltage",
            data=read_max_high_voltage(file),
            description="The maximum high voltage used by the Verasonics system.",
            unit="V",
        )


    zea_from_matlab_raw(
        "C:/path/to/raw_data.mat",
        "C:/path/to/output.hdf5",
        [read_high_voltage_func],
    )
"""  # noqa: E501

import argparse
import os
import sys
import tkinter as tk
import traceback
from pathlib import Path
from tkinter import filedialog

import h5py
import numpy as np

from zea import log
from zea.data.data_format import DatasetElement, generate_zea_dataset
from zea.ops import LogCompress, Normalize
from zea.utils import strtobool

_VERASONICS_TO_ZEA_PROBE_NAMES = {
    "L11-4v": "verasonics_l11_4v",
    "L11-5v": "verasonics_l11_5v",
}


def dereference_index(file, dataset, index, event=None, subindex=None):
    """Get the element at the given index from the dataset, dereferencing it if
    necessary.

    MATLAB stores items in struct array differently depending on the size. If the size
    is 1, the item is stored as a regular dataset. If the size is larger, the item is
    stored as a dataset of references to the actual data.

    This function dereferences the dataset if it is a reference. Otherwise, it returns
    the dataset.

    Args:
        file (h5py.File): The file to read the dataset from.
        dataset (h5py.Dataset): The dataset to read the element from.
        index (int): The index of the element to read.
        event (int, optional): The event index. Usually we store each event in the
            second dimension of the dataset. Defaults to None in this case we assume
            that there is only a single event.
        subindex (slice, optional): The subindex of the element to read after
            referencing the actual data. Defaults to None. In this case, all the data
            is returned.
    """
    if isinstance(dataset.fillvalue, h5py.h5r.Reference):
        if event is not None:
            reference = dataset[index, event]
        else:
            reference = dataset[index, 0]
        if subindex is None:
            return file[reference][:]
        else:
            return file[reference][subindex]
    else:
        if index > 0:
            log.warning(
                f"index {index} is not a reference. You are probably "
                "incorrectly indexing a dataset."
            )
        return dataset


def get_reference_size(dataset):
    """Get the size of a reference dataset."""
    if isinstance(dataset.fillvalue, h5py.h5r.Reference):
        return len(dataset)
    else:
        return 1


def decode_string(dataset):
    """Decode a string dataset."""
    return "".join([chr(c) for c in dataset.squeeze()])


def read_probe_geometry(file):
    """
    Read the probe geometry from the file.

    Args:
        file (h5py.File): The file to read the probe geometry from. (The file should
            be opened in read mode.)

    Returns:
        probe_geometry (np.ndarray): The probe geometry of shape (n_el, 3).
    """
    # Read the probe geometry from the file
    probe_geometry = file["Trans"]["ElementPos"][:3, :]

    # Transpose the probe geometry to have the shape (n_el, 3)
    probe_geometry = probe_geometry.T

    # Read the unit
    unit = decode_string(file["Trans"]["units"][:])

    # Convert the probe geometry to meters
    if unit == "mm":
        probe_geometry = probe_geometry / 1000
    else:
        wavelength = read_wavelength(file)
        probe_geometry = probe_geometry * wavelength

    return probe_geometry


def read_wavelength(file):
    """Reads the wavelength from the file.

    Args:
        `file` (`h5py.File`): The file to read the wavelength from. (The file should be
            opened in read mode.)

    Returns:
        `wavelength` (`float`): The wavelength of the probe.
    """
    center_frequency = read_probe_center_frequency(file)
    sound_speed = read_sound_speed(file)
    wavelength = sound_speed / center_frequency
    return wavelength


def read_transmit_events(file, event=None, frames="all"):
    """Read the events from the file and finds the order in which transmits and receives
    appear in the events.

    Args:
        file (h5py.File): The file to read the events from.
            The file should be opened in read mode.
        event (int, optional): The event index. Defaults to None.
        frames (str or list, optional): The frames to read. Defaults to "all".

    Returns:
        tuple: (tx_order, rcv_order, time_to_next_acq)
            tx_order (list): The order in which the transmits appear in the events.
            rcv_order (list): The order in which the receives appear in the events.
            time_to_next_acq (np.ndarray): The time to next acquisition of shape (n_acq, n_tx).
    """

    num_events = file["Event"]["info"].shape[0]

    # In the Verasonics the transmits may not be in order in the TX structure and a
    # transmit might be reused. Therefore, we need to keep track of the order in which
    # the transmits appear in the Events.
    tx_order = []
    rcv_order = []
    time_to_next_acq = []

    frame_indices = get_frame_indices(file, frames)

    for i in range(num_events):
        # Get the tx
        event_tx = dereference_index(file, file["Event"]["tx"], i)
        event_tx = int(event_tx.item())

        # Get the rcv
        event_rcv = dereference_index(file, file["Event"]["rcv"], i)
        event_rcv = int(event_rcv.item())

        if not bool(event_tx) == bool(event_rcv):
            log.warning(
                "Events should have both a transmit and a receive or neither. "
                f"Event {i} has a transmit but no receive or vice versa."
            )

        if not event_tx:
            continue

        # Subtract one to make the indices 0-based
        event_tx -= 1
        event_rcv -= 1

        # Check in the Receive structure if this is still the first frame
        framenum_ref = file["Receive"]["framenum"][event_rcv, 0]
        framenum = file[framenum_ref][:].item()

        # Only add the event to the list if it is the first frame since we assume
        # that all frames have the same transmits and receives
        if framenum == 1:
            # Add the event to the list
            tx_order.append(event_tx)
            rcv_order.append(event_rcv)

        # Read the time_to_next_acq
        seq_control_indices = dereference_index(file, file["Event"]["seqControl"], i)

        for seq_control_index in seq_control_indices:
            seq_control_index = int(seq_control_index.item() - 1)
            seq_control = dereference_index(file, file["SeqControl"]["command"], seq_control_index)
            # Decode the seq_control int array into a string
            seq_control = decode_string(seq_control)
            if seq_control == "timeToNextAcq":
                value = dereference_index(
                    file, file["SeqControl"]["argument"], seq_control_index
                ).item()
                value = value * 1e-6
                time_to_next_acq.append(value)

    n_tx = len(tx_order)
    time_to_next_acq = np.array(time_to_next_acq)
    time_to_next_acq = np.reshape(time_to_next_acq, (-1, n_tx))

    if event is not None:
        time_to_next_acq = time_to_next_acq[event]
        time_to_next_acq = np.expand_dims(time_to_next_acq, axis=0)

    time_to_next_acq = time_to_next_acq[frame_indices]

    return tx_order, rcv_order, time_to_next_acq


def read_t0_delays_apod(file, tx_order, event=None):
    """
    Read the t0 delays and apodization from the file.

    Args:
        file (h5py.File): The file to read the t0 delays from. (The file should be
            opened in read mode.)

    Returns:
        t0_delays (np.ndarray): The t0 delays of shape (n_tx, n_el).
        apod (np.ndarray): The apodization of shape (n_el,).
    """

    t0_delays_list = []
    tx_apodizations_list = []

    wavelength = read_wavelength(file)
    sound_speed = read_sound_speed(file)

    for n in tx_order:
        # Get column vector of t0_delays
        if event is None:
            t0_delays = dereference_index(file, file["TX"]["Delay"], n)
        else:
            t0_delays = dereference_index(file, file["TX_Agent"]["Delay"], n, event)
        # Turn into 1d array
        t0_delays = t0_delays[:, 0]

        t0_delays_list.append(t0_delays)

        # Get column vector of apodizations
        if event is None:
            tx_apodizations = dereference_index(file, file["TX"]["Apod"], n)
        else:
            tx_apodizations = dereference_index(file, file["TX_Agent"]["Apod"], n, event)
        # Turn into 1d array
        tx_apodizations = tx_apodizations[:, 0]
        tx_apodizations_list.append(tx_apodizations)

    t0_delays = np.stack(t0_delays_list, axis=0)
    apodizations = np.stack(tx_apodizations_list, axis=0)

    # Convert the t0_delays to meters
    t0_delays = t0_delays * wavelength / sound_speed

    return t0_delays, apodizations


def read_sampling_frequency(file):
    """
    Read the sampling frequency from the file.

    Args:
        file (h5py.File): The file to read the sampling frequency from. (The file
            should be opened in read mode.)

    Returns:
        sampling_frequency (float): The sampling frequency.
    """
    # Read the sampling frequency from the file
    adc_rate = dereference_index(file, file["Receive"]["decimSampleRate"], 0)

    # The Vantage NXT has renamed this field to sampleSkip
    if "quadDecim" in file["Receive"]:
        quaddecim = dereference_index(file, file["Receive"]["quadDecim"], 0)
    else:
        quaddecim = 1.0

    sampling_frequency = adc_rate / quaddecim * 1e6

    return sampling_frequency[0, 0]


def read_waveforms(file, tx_order, event=None):
    """
    Read the waveforms from the file.

    Args:
        file (h5py.File): The file to read the waveforms from. (The file should be
            opened in read mode.)

    Returns:
        waveforms (np.ndarray): The waveforms of shape (n_tx, n_samples).
    """
    waveforms_one_way_list = []
    waveforms_two_way_list = []

    # Read all the waveforms from the file
    n_tw = get_reference_size(file["TW"]["Wvfm1Wy"])
    for n in range(n_tw):
        # Get the row vector of the 1-way waveform
        waveform_one_way = dereference_index(file, file["TW"]["Wvfm1Wy"], n)[:]
        # Turn into 1d array
        waveform_one_way = waveform_one_way[0, :]

        # Get the row vector of the 2-way waveform
        waveform_two_way = dereference_index(file, file["TW"]["Wvfm2Wy"], n)[:]
        # Turn into 1d array
        waveform_two_way = waveform_two_way[0, :]

        waveforms_one_way_list.append(waveform_one_way)
        waveforms_two_way_list.append(waveform_two_way)

    tx_waveform_indices = []

    for n in tx_order:
        # Read the waveform
        if event is None:
            waveform_index = dereference_index(file, file["TX"]["waveform"], n)[:]
        else:
            waveform_index = dereference_index(file, file["TX_Agent"]["waveform"], n, event)[:]
        # Subtract one to make the indices 0-based
        waveform_index -= 1
        # Turn into integer
        waveform_index = int(waveform_index.item())
        tx_waveform_indices.append(waveform_index)

    return tx_waveform_indices, waveforms_one_way_list, waveforms_two_way_list


def read_polar_angles(file, tx_order, event=None):
    """
    Read the polar angles from the file.

    Args:
        file (h5py.File): The file to read the polar angles from. (The file should
            be opened in read mode.)

    Returns:
        polar_angles (np.ndarray): The polar angles of shape (n_tx,).
    """
    polar_angles_list = []

    for n in tx_order:
        # Read the polar angle
        if event is None:
            polar_angle = dereference_index(file, file["TX"]["Steer"], n)[:]
        else:
            polar_angle = dereference_index(file, file["TX_Agent"]["Steer"], n, event)[:]
        # Turn into 1d array
        polar_angle = polar_angle[0, 0]

        polar_angles_list.append(polar_angle)

    polar_angles = np.stack(polar_angles_list, axis=0)

    return polar_angles


def read_azimuth_angles(file, tx_order, event=None):
    """
    Read the azimuth angles from the file.

    Args:
        file (h5py.File): The file to read the azimuth angles from. (The file should
            be opened in read mode.)

    Returns:
        azimuth_angles (np.ndarray): The azimuth angles of shape (n_tx,).
    """
    azimuth_angles_list = []

    for n in tx_order:
        # Read the azimuth angle
        if event is None:
            azimuth_angle = dereference_index(file, file["TX"]["Steer"], n)[:]
        else:
            azimuth_angle = dereference_index(file, file["TX_Agent"]["Steer"], n, event)[:]
        # Turn into 1d array
        azimuth_angle = azimuth_angle[1, 0]

        azimuth_angles_list.append(azimuth_angle)

    azimuth_angles = np.stack(azimuth_angles_list, axis=0)

    return azimuth_angles


def read_raw_data(file, event=None, frames="all"):
    """
    Read the raw data from the file.

    Args:
        file (h5py.File): The file to read the raw data from. (The file should be
            opened in read mode.)

    Returns:
        raw_data (np.ndarray): The raw data of shape (n_rcv, n_samples).
    """

    # Get the number of axial samples
    start_sample = dereference_index(file, file["Receive"]["startSample"], 0).item()
    end_sample = dereference_index(file, file["Receive"]["endSample"], 0).item()
    n_ax = int(end_sample - start_sample + 1)

    # Obtain the number of transmit events per frame
    tx_order, _, _ = read_transmit_events(file)
    n_tx = len(tx_order)

    # Read the raw data from the file
    if event is None:
        raw_data = dereference_index(file, file["RcvData"], 0)
    else:
        # for now we only index frames as events
        raw_data = dereference_index(file, file["RcvData"], 0, subindex=event)
        raw_data = np.expand_dims(raw_data, axis=0)

    frame_indices = get_frame_indices(file, frames)

    raw_data = raw_data[frame_indices]

    raw_data = raw_data[:, :, : n_ax * n_tx]

    raw_data = raw_data.reshape((raw_data.shape[0], raw_data.shape[1], n_tx, -1))

    raw_data = np.transpose(raw_data, (0, 2, 3, 1))

    # Add channel dimension
    raw_data = raw_data[..., None]

    return raw_data


def read_probe_center_frequency(file):
    """Reads the center frequency of the probe from the file.

    Args:
        file (h5py.File): The file to read the center frequency from. (The file
            should be opened in read mode.)

    Returns:
        center_frequency (float): The center frequency of the probe.
    """
    center_frequency = file["Trans"]["frequency"][0, 0] * 1e6
    return center_frequency


def read_sound_speed(file):
    """Reads the speed of sound from the file.

    Args:
        file (h5py.File): The file to read the speed of sound from. (The file
            should be opened in read mode.)

    Returns:
        sound_speed (float): The speed of sound.
    """

    sound_speed = file["Resource"]["Parameters"]["speedOfSound"][0, 0].item()
    return sound_speed


def read_initial_times(file, rcv_order, sound_speed):
    """Reads the initial times from the file.

    Args:
        file (h5py.File): The file to read the initial times from. (The file should
            be opened in read mode.)
        rcv_order (list): The order in which the receives appear in the events.
        wavelength (float): The wavelength of the probe.
        sound_speed (float): The speed of sound.

    Returns:
        initial_times (np.ndarray): The initial times of shape (n_rcv,).
    """
    wavelength = read_wavelength(file)
    initial_times = []
    for n in rcv_order:
        start_depth = dereference_index(file, file["Receive"]["startDepth"], n).item()

        initial_times.append(2 * start_depth * wavelength / sound_speed)

    return np.array(initial_times).astype(np.float32)


def read_probe_name(file):
    """Reads the name of the probe from the file.

    Args:
        file (h5py.File): The file to read the name of the probe from. (The file
            should be opened in read mode.)

    Returns:
        probe_name (str): The name of the probe.
    """
    probe_name = file["Trans"]["name"][:]
    probe_name = decode_string(probe_name)
    # Translates between verasonics probe names and zea probe names
    if probe_name in _VERASONICS_TO_ZEA_PROBE_NAMES:
        probe_name = _VERASONICS_TO_ZEA_PROBE_NAMES[probe_name]
    else:
        log.warning(
            f"Probe name {probe_name} is not in the list of known probes. "
            "Please add it to the _VERASONICS_TO_ZEA_PROBE_NAMES dictionary. "
            "Falling back to generic probe."
        )
        probe_name = "generic"

    return probe_name


def read_focus_distances(file, tx_order, event=None):
    """Reads the focus distances from the file.

    Args:
        file (h5py.File): The file to read the focus distances from. (The file
            should be opened in read mode.)
        tx_order (list): The order in which the transmits appear in the events.

    Returns:
        focus_distances (list): The focus distances.
    """
    focus_distances = []
    for n in tx_order:
        if event is None:
            focus_distance = dereference_index(file, file["TX"]["focus"], n)[0, 0]
        else:
            focus_distance = dereference_index(file, file["TX_Agent"]["focus"], n, event)[0, 0]
        focus_distances.append(focus_distance)
    return np.array(focus_distances)


def _probe_geometry_is_ordered_ula(probe_geometry):
    """Checks if the probe geometry is ordered as a uniform linear array (ULA)."""
    diff_vec = probe_geometry[1:] - probe_geometry[:-1]
    return np.isclose(diff_vec, diff_vec[0]).all()


def planewave_focal_distance_to_inf(focus_distances, t0_delays, tx_apodizations, probe_geometry):
    """Detects plane wave transmits and sets the focus distance to infinity.

    Args:
        focus_distances (np.ndarray): The focus distances of shape (n_tx,).
        t0_delays (np.ndarray): The t0 delays of shape (n_tx, n_el).
        tx_apodizations (np.ndarray): The apodization of shape (n_tx, n_el).

    Returns:
        focus_distances (np.ndarray): The focus distances of shape (n_tx,).

    Note:
        This function assumes that the probe_geometry is a 1d uniform linear array.
        If not it will warn and return.
    """
    if not _probe_geometry_is_ordered_ula(probe_geometry):
        log.warning(
            "The probe geometry is not ordered as a uniform linear array. "
            "Focal distances are not set to infinity for plane waves."
        )
        return focus_distances

    for tx in range(focus_distances.size):
        mask_active = np.abs(tx_apodizations[tx]) > 0
        if np.sum(mask_active) < 2:
            continue
        t0_delays_active = t0_delays[tx][mask_active]

        # If the t0_delays all have the same offset, we assume it is a plane wave
        if np.std(np.diff(t0_delays_active)) < 1e-16:
            focus_distances[tx] = np.inf

    return focus_distances


def read_bandwidth_percent(file):
    """Reads the bandwidth percent from the file.

    Args:
        file (h5py.File): The file to read the bandwidth percent from. (The file
            should be opened in read mode.)

    Returns:
        bandwidth_percent (int): The bandwidth percent.
    """
    bandwidth_percent = dereference_index(file, file["Receive"]["sampleMode"], 0)
    bandwidth_percent = decode_string(bandwidth_percent)
    bandwidth_percent = int(bandwidth_percent[2:-2])
    return bandwidth_percent


def read_lens_correction(file):
    """Reads the lens correction from the file.

    Args:
        `file` (`h5py.File`): The file to read the lens correction from. (The file
            should be opened in read mode.)

    Returns:
        `lens_correction` (`np.ndarray`): The lens correction.
    """
    lens_correction = file["Trans"]["lensCorrection"][0, 0].item()
    return lens_correction


def read_tgc_gain_curve(file):
    """Reads the TGC gain curve from the file.

    Parameters
    ----------
    file : h5py.File
        The file to read the TGC gain curve from. (The file should be opened in read
        mode.)

    Returns
    -------
    np.ndarray
        The TGC gain curve of shape `(n_ax,)`.
    """

    gain_curve = file["TGC"]["Waveform"][:][:, 0]

    # Normalize the gain_curve to [0, 40]dB
    gain_curve = gain_curve / 1023 * 40

    # The gain curve is sampled at 800ns (See Verasonics documentation for details.
    # Specifically the tutorial sequence programming)
    gain_curve_sampling_period = 800e-9

    # Define the time axis for the gain curve
    t_gain_curve = np.arange(gain_curve.size) * gain_curve_sampling_period

    # Read the number of axial samples
    start_sample = dereference_index(file, file["Receive"]["startSample"], 0).item()
    end_sample = dereference_index(file, file["Receive"]["endSample"], 0).item()
    n_ax = int(end_sample - start_sample + 1)

    # Read the sampling frequency
    sampling_frequency = read_sampling_frequency(file)

    # Define the time axis for the axial samples
    t_samples = np.arange(n_ax) / sampling_frequency

    # Interpolate the gain_curve to the number of axial samples
    gain_curve = np.interp(t_samples, t_gain_curve, gain_curve)

    # The gain_curve gains are in dB, so we need to convert them to linear scale
    gain_curve = 10 ** (gain_curve / 20)

    return gain_curve


def read_image_data_p(file, event=None, frames="all"):
    """Reads the image data from the file.

    Args:
        `file` (`h5py.File`): The file to read the image data from. (The file should be
            opened in read mode.)

    Returns:
        `image_data` (`np.ndarray`): The image data.
    """
    # Check if the file contains image data
    if "ImgDataP" not in file:
        return None

    frame_indices = get_frame_indices(file, frames)

    # Get the dataset reference
    image_data_ref = file["ImgDataP"][0, 0]
    # Dereference the dataset
    if event is None:
        image_data = file[image_data_ref][:]
    else:
        image_data = file[image_data_ref][event]
        image_data = np.expand_dims(image_data, axis=0)

    # Get the relevant dimensions
    image_data = image_data[:, 0, :, :]

    # Convert to [-60, 0] dB range based on min and max values
    normalize = Normalize(output_range=(0, 1), input_range=None)
    log_compress = LogCompress()

    image_data = normalize(data=image_data)["data"]
    image_data = log_compress(data=image_data, dynamic_range=(-np.inf, 0))["data"]

    # Reshape so that [n_frames, n_samples, n_lines]
    image_data = np.transpose(image_data, (0, 2, 1))

    image_data = image_data[frame_indices]

    return image_data


def read_probe_element_width(file):
    """Reads the element width from the file.

    Args:
        file (h5py.File): The file to read the element width from.
            The file should be opened in read mode.

    Returns:
        float: The element width.
    """
    element_width = file["Trans"]["elementWidth"][:][0, 0]

    # Read the unit
    unit = decode_string(file["Trans"]["units"][:])

    # Convert the probe element width to meters
    if unit == "mm":
        element_width = element_width / 1000
    else:
        wavelength = read_wavelength(file)
        element_width = element_width * wavelength

    return element_width


def read_verasonics_file(file, event=None, additional_functions=None, frames="all"):
    """Reads data from a .mat Verasonics output file.

    Args:
        file (h5py.File): The file to read the data from. (The file should be opened in
            read mode.)
        event (int, optional): The event index. Defaults to None in this case we assume
            the data file is stored without event structure.
        additional_functions (list, optional): A list of functions that read additional
            data from the file. Each function should take the file as input and return a
            `DatasetElement`. Defaults to None.
    """

    probe_geometry = read_probe_geometry(file)

    # same for all events
    tx_order, rcv_order, time_to_next_transmit = read_transmit_events(file, frames=frames)
    sampling_frequency = read_sampling_frequency(file)
    bandwidth_percent = read_bandwidth_percent(file)
    center_frequency = read_probe_center_frequency(file)
    sound_speed = read_sound_speed(file)
    initial_times = read_initial_times(file, rcv_order, sound_speed)
    probe_name = read_probe_name(file)
    tgc_gain_curve = read_tgc_gain_curve(file)
    element_width = read_probe_element_width(file)

    # these are capable of handling multiple events
    raw_data = read_raw_data(file, event, frames=frames)
    image = read_image_data_p(file, event, frames=frames)

    polar_angles = read_polar_angles(file, tx_order, event)
    azimuth_angles = read_azimuth_angles(file, tx_order, event)
    t0_delays, tx_apodizations = read_t0_delays_apod(file, tx_order, event)
    focus_distances = read_focus_distances(file, tx_order, event)

    tx_waveform_indices, waveforms_one_way_list, waveforms_two_way_list = read_waveforms(
        file, tx_order, event
    )
    focus_distances = planewave_focal_distance_to_inf(
        focus_distances, t0_delays, tx_apodizations, probe_geometry
    )

    # If the data is captured in BS100BW mode or BS50BW mode, the data is stored in
    # as complex IQ data and the sampling frequency is halved.
    if bandwidth_percent in (50, 100):
        raw_data = np.concatenate(
            (
                raw_data[:, :, 0::2, :, :],
                -raw_data[:, :, 1::2, :, :],
            ),
            axis=-1,
        )
        # Two sequential samples are interpreted as a single complex sample
        # Therefore, we need to halve the sampling frequency
        sampling_frequency = sampling_frequency / 2

        # We have halved the number of samples, so we need to halve the number
        # of samples in the gain curve as well
        tgc_gain_curve = tgc_gain_curve[0::2]

    lens_correction = read_lens_correction(file)
    if event is None:
        group_name = "scan"
    else:
        group_name = f"event_{event}/scan"

    el_lens_correction = DatasetElement(
        group_name=group_name,
        dataset_name="lens_correction",
        data=lens_correction,
        description=(
            "The lens correction value used by Verasonics. This value is the "
            "additional path length in wavelength that the lens introduces. "
            "(This disregards refraction.)"
        ),
        unit="wavelengths",
    )

    additional_elements = []
    if additional_functions is not None:
        for additional_function in additional_functions:
            additional_elements.append(additional_function(file))

    data = {
        "probe_geometry": probe_geometry,
        "time_to_next_transmit": time_to_next_transmit,
        "t0_delays": t0_delays,
        "tx_apodizations": tx_apodizations,
        "sampling_frequency": sampling_frequency,
        "polar_angles": polar_angles,
        "azimuth_angles": azimuth_angles,
        "bandwidth_percent": bandwidth_percent,
        "raw_data": raw_data,
        "image": image,
        "center_frequency": center_frequency,
        "sound_speed": sound_speed,
        "initial_times": initial_times,
        "probe_name": probe_name,
        "focus_distances": focus_distances,
        "tx_waveform_indices": tx_waveform_indices,
        "waveforms_one_way": waveforms_one_way_list,
        "waveforms_two_way": waveforms_two_way_list,
        "tgc_gain_curve": tgc_gain_curve,
        "element_width": element_width,
        "additional_elements": [el_lens_correction, *additional_elements],
    }

    return data


def get_frame_indices(file, frames):
    """Creates a numpy array of frame indices from the file and the frames argument.

    Args:
        file (h5py.File): The file to read the frame indices from.
        frames (str): The frames argument. This can be "all" or a list of frame indices.

    Returns:
        frame_indices (np.ndarray): The frame indices.
    """
    # Read the number of frames from the file
    n_frames = int(file["Resource"]["RcvBuffer"]["numFrames"][0][0])

    if isinstance(frames, str) and frames == "all":
        # Create an array of all frame-indices
        frame_indices = np.arange(n_frames)
    else:
        frame_indices = np.array(frames)
        frame_indices.sort()

    if np.any(frame_indices >= n_frames):
        log.error(
            f"Frame indices {frame_indices} are out of bounds. "
            f"The file contains {n_frames} frames. "
            f"Using only the indices that are within bounds."
        )
        # Remove out of bounds indices
        frame_indices = frame_indices[frame_indices < n_frames]

    return frame_indices


def zea_from_matlab_raw(input_path, output_path, additional_functions=None, frames="all"):
    """Converts a Verasonics matlab raw file to the zea format. The MATLAB file
    should be created using the `save_raw` function and be stored in "v7.3" format.

    Args:
        input_path (str): The path to the input file (.mat file).
        output_path (str): The path to the output file (.hdf5 file).
        additional_functions (list, optional): A list of functions that read additional
            data from the file. Each function should take the file as input and return a
            `DatasetElement`. Defaults to None.
        frames (str or list of int, optional): The frames to add to the file. This can be
            a list of integers, a range of integers (e.g. 4-8), or 'all'. Defaults to
            'all'.
    """
    # Create the output directory if it does not exist
    input_path = Path(input_path)
    output_path = Path(output_path)

    output_path.parent.mkdir(parents=True, exist_ok=True)

    assert input_path.is_file(), log.error(f"Input file {log.yellow(input_path)} does not exist.")

    # Load the data
    with h5py.File(input_path, "r") as file:
        if "TX_Agent" in file:
            active_keys = file["TX_Agent"].keys()
            log.info(
                f"Found active imaging data with {len(active_keys)} events. "
                "Will convert and save all parameters for each event separately."
            )
            num_events = set(file["TX_Agent"][key].shape[-1] for key in active_keys)
            assert len(num_events) == 1, (
                "All TX_Agent entries should have the same number of events."
            )
            num_events = num_events.pop()

            # loop over TX_Agent entries and overwrite TX each time
            data = {}
            for event in range(num_events):
                data[event] = read_verasonics_file(
                    file,
                    event=event,
                    additional_functions=additional_functions,
                )

            # convert dict of events to dict of lists
            data = {key: [data[event][key] for event in data] for key in data[0]}
            description = ["Verasonics data with multiple events"] * num_events
            # Generate the zea dataset
            generate_zea_dataset(
                path=output_path,
                **data,
                event_structure=True,
                description=description,
            )

        else:
            # Here we call al the functions to read the data from the file
            data = read_verasonics_file(
                file, additional_functions=additional_functions, frames=frames
            )

            # Generate the zea dataset
            generate_zea_dataset(path=output_path, **data, description="Verasonics data")

    log.success(f"Converted {log.yellow(input_path)} to {log.yellow(output_path)}")


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Convert Verasonics matlab raw files to the zea format."
        "Example usage: python zea/data/convert/matlab.py raw_file.mat output.hdf5 --frames 1-5 7"
    )
    parser.add_argument(
        "input_path",
        default=None,
        type=str,
        nargs="?",
        help="The path to a file or directory containing raw Verasonics data.",
    )

    parser.add_argument(
        "output_path",
        default=None,
        type=str,
        nargs="?",
        help="The path to the output file or directory.",
    )

    parser.add_argument(
        "--frames",
        default=["all"],
        type=str,
        nargs="+",
        help="The frames to add to the file. This can be a list of integers, a range "
        "of integers (e.g. 4-8), or 'all'.",
    )

    return parser.parse_args()


def get_answer(prompt, additional_options=None):
    """Get a yes or no answer from the user. There is also the option to provide
    additional options. In case yes or no is selected, the function returns a boolean.
    In case an additional option is selected, the function returns the selected option
    as a string.

    Args:
        prompt (str): The prompt to show the user.
        additional_options (list, optional): Additional options to show the user.
            Defaults to None.

    Returns:
        str: The user's answer.
    """
    while True:
        answer = input(prompt)
        try:
            bool_answer = strtobool(answer)
            return bool_answer
        except ValueError:
            if additional_options is not None and answer in additional_options:
                return answer
        log.warning("Invalid input.")


if __name__ == "__main__":
    args = parse_args()

    # Variable to indicate what to do with existing files.
    # Is set by the user in case these are found.
    existing_file_policy = None

    if args.input_path is None:
        log.info("Select a directory containing Verasonics matlab raw files.")
        # Create a Tkinter root window
        try:
            root = tk.Tk()
            root.withdraw()
            # Prompt the user to select a file or directory
            selected_path = filedialog.askdirectory()
        except Exception as e:
            raise ValueError(
                log.error(
                    "Failed to open a file dialog (possibly in headless state). "
                    "Please provide a path as an argument. "
                )
            ) from e
    else:
        selected_path = args.input_path

    # Exit when no path is selected
    if not selected_path:
        log.error("No path selected.")
        sys.exit()
    else:
        selected_path = Path(selected_path)

    selected_path_is_directory = os.path.isdir(selected_path)

    # Set the output path to be next to the input directory with _zea appended
    # to the name
    if args.output_path is None:
        if selected_path_is_directory:
            output_path = selected_path.parent / (Path(selected_path).name + "_zea")
        else:
            output_path = str(selected_path.with_suffix("")) + "_zea.hdf5"
            output_path = Path(output_path)
    else:
        output_path = Path(args.output_path)
        if selected_path.is_file() and output_path.suffix not in (".hdf5", ".h5"):
            log.error(
                "When converting a single file, the output path should have the .hdf5 "
                "or .h5 extension."
            )
            sys.exit()
        elif selected_path.is_dir() and output_path.is_file():
            log.error("When converting a directory, the output path should be a directory.")
            sys.exit()
        #
        if output_path.is_dir() and not selected_path_is_directory:
            output_path = output_path / (selected_path.name + "_zea.hdf5")

    log.info(f"Selected path: {log.yellow(selected_path)}")

    # Parse frames
    frames = args.frames
    if frames[0] == "all":
        frames = "all"
    else:
        indices = set()
        for frame in frames:
            if "-" in frame:
                start, end = frame.split("-")
                indices.update(range(int(start), int(end) + 1))
            else:
                indices.add(int(frame))
        frames = list(indices)
        frames.sort()
    # Do the conversion of a single file
    if not selected_path_is_directory:
        if output_path.is_file():
            answer = get_answer(
                f"File {log.yellow(output_path)} exists. Overwrite?"
                "\n\ty\t - Overwrite"
                "\n\tn\t - Skip"
                "\nAnswer: "
            )
            if answer is True:
                log.warning(f"{selected_path} exists. Deleting...")
                output_path.unlink(missing_ok=False)
            else:
                log.info("Aborting...")
                sys.exit()
        zea_from_matlab_raw(selected_path, output_path, frames=frames)
    else:
        # Continue with the rest of your code...
        for root, dirs, files in os.walk(selected_path):
            for mat_file in files:
                # Skip non-mat files
                if not mat_file.endswith(".mat"):
                    continue

                log.info(f"Found raw data file {log.yellow(mat_file)}")

                # Convert the file to a Path object
                mat_file = Path(mat_file)

                # Construct the output path
                relative_path = (Path(root) / Path(mat_file)).relative_to(selected_path)
                file_output_path = output_path / (relative_path.with_suffix(".hdf5"))

                full_path = selected_path / relative_path

                # Handle existing files
                if file_output_path.is_file():
                    if existing_file_policy is None:
                        answer = get_answer(
                            f"File {log.yellow(file_output_path)} exists. Overwrite?"
                            "\n\ty\t - Overwrite"
                            "\n\tn\t - Skip"
                            "\n\tya\t - Overwrite all existing files"
                            "\n\tna\t - Skip all existing files"
                            "\nAnswer: ",
                            additional_options=("ya", "na"),
                        )
                        if answer == "ya":
                            existing_file_policy = "overwrite"
                        elif answer == "na":
                            existing_file_policy = "skip"
                            continue

                    if existing_file_policy == "skip" or answer is False:
                        log.info("Skipping...")
                        continue

                    if existing_file_policy == "overwrite" or answer is True:
                        log.warning(f"{log.yellow(full_path)} exists. Deleting...")
                        file_output_path.unlink(missing_ok=False)

                try:
                    zea_from_matlab_raw(full_path, file_output_path, frames=frames)
                except Exception:
                    # Print error message without raising it
                    log.error(f"Failed to convert {mat_file}")
                    # Print stacktrace
                    traceback.print_exc()

                    continue



"""Pixel grid calculation for ultrasound beamforming."""

import numpy as np

from zea import log

eps = 1e-10


def get_grid(
    xlims,
    zlims,
    Nx: int,
    Nz: int,
    sound_speed,
    center_frequency,
    pixels_per_wavelength,
    verbose=False,
):
    """Creates a pixelgrid based on scan class parameters."""

    if Nx and Nz:
        grid = cartesian_pixel_grid(xlims, zlims, Nx=int(Nx), Nz=int(Nz))
    else:
        wvln = sound_speed / center_frequency
        dx = wvln / pixels_per_wavelength
        dz = dx
        grid = cartesian_pixel_grid(xlims, zlims, dx=dx, dz=dz)
        if verbose:
            print(
                f"Pixelgrid was set automatically to Nx: {grid.shape[1]}, Nz: {grid.shape[0]}, "
                f"using {pixels_per_wavelength} pixels per wavelength."
            )
    return grid


def check_for_aliasing(scan):
    """Checks if the scan class parameters will cause spatial aliasing due to a too low pixel
    density. If so, a warning is printed with a suggestion to increase the pixel density by either
    increasing the number of pixels, or decreasing the pixel spacing, depending on which parameter
    was set by the user."""
    wvln = scan.sound_speed / scan.center_frequency
    dx = wvln / scan.pixels_per_wavelength
    dz = dx

    width = scan.xlims[1] - scan.xlims[0]
    depth = scan.zlims[1] - scan.zlims[0]

    if scan.Nx and scan.Nz:
        if width / scan.Nx > wvln / 2:
            log.warning(
                f"width/Nx = {width / scan.Nx:.7f} < wvln/2 = {wvln / 2}. "
                f"Consider increasing scan.Nx to {int(np.ceil(width / (wvln / 2)))} or more."
            )
        if depth / scan.Nz > wvln / 2:
            log.warning(
                f"depth/Nz = {depth / scan.Nz:.7f} < wvln/2 = {wvln / 2:.7f}. "
                f"Consider increasing scan.Nz to {int(np.ceil(depth / (wvln / 2)))} or more."
            )
    else:
        if dx > wvln / 2:
            log.warning(
                f"dx = {dx:.7f} > wvln/2 = {wvln / 2:.7f}. "
                f"Consider increasing scan.pixels_per_wavelength to 2 or more"
            )
        if dz > wvln / 2:
            log.warning(
                f"dz = {dz:.7f} > wvln/2 = {wvln / 2:.7f}. "
                f"Consider increasing scan.pixels_per_wavelength to 2 or more"
            )


def cartesian_pixel_grid(xlims, zlims, Nx=None, Nz=None, dx=None, dz=None):
    """Generate a Cartesian pixel grid based on input parameters.

    Args:
        xlims (tuple): Azimuthal limits of pixel grid ([xmin, xmax])
        zlims (tuple): Depth limits of pixel grid ([zmin, zmax])
        Nx (int): Number of azimuthal pixels, overrides dx and dz parameters
        Nz (int): Number of depth pixels, overrides dx and dz parameters
        dx (float): Pixel spacing in azimuth
        dz (float): Pixel spacing in depth

    Raises:
        ValueError: Either Nx and Nz or dx and dz must be defined.

    Returns:
        grid (np.ndarray): Pixel grid of size (nz, nx, 3) in
            Cartesian coordinates (x, y, z)
    """
    assert (bool(Nx) and bool(Nz)) ^ (bool(dx) and bool(dz)), (
        "Either Nx and Nz or dx and dz must be defined."
    )

    # Determine the grid spacing
    if Nx is not None and Nz is not None:
        x = np.linspace(xlims[0], xlims[1] + eps, Nx)
        z = np.linspace(zlims[0], zlims[1] + eps, Nz)
    elif dx is not None and dz is not None:
        sign = np.sign(xlims[1] - xlims[0])
        x = np.arange(xlims[0], xlims[1] + eps, sign * dx)
        z = np.arange(zlims[0], zlims[1] + eps, sign * dz)
    else:
        raise ValueError("Either Nx and Nz or dx and dz must be defined.")

    # Create the pixel grid
    z_grid, x_grid = np.meshgrid(z, x, indexing="ij")
    y_grid = 0 * x_grid  # Assume y = 0
    grid = np.stack((x_grid, y_grid, z_grid), axis=-1)
    return grid


def radial_pixel_grid(rlims, dr, oris, dirs):
    """Generate a focused pixel grid based on input parameters.

    To accommodate the multitude of ways of defining a focused transmit grid, we define
    pixel "rays" or "lines" according to their origins (oris) and directions (dirs).
    The position along the ray is defined by its limits (rlims) and spacing (dr).

    Args:
        rlims (tuple): Radial limits of pixel grid ([rmin, rmax])
        dr (float): Pixel spacing in radius
        oris (np.ndarray): Origin of each ray in Cartesian coordinates (x, y, z)
            with shape (nrays, 3)
        dirs (np.ndarray): Steering direction of each ray in azimuth, in units of
            radians (nrays, 2)

    Returns:
        grid (np.ndarray): Pixel grid of size (nr, nrays, 3) in
            Cartesian coordinates (x, y, z), with nr being the number of radial pixels.
    """
    # Get focusing positions in rho-theta coordinates
    r = np.arange(rlims[0], rlims[1] + eps, dr)  # Depth rho
    t = dirs[:, 0]  # Use azimuthal angle theta (ignore elevation angle)
    tt, rr = np.meshgrid(t, r, indexing="ij")

    # Convert the focusing grid to Cartesian coordinates
    xx = rr * np.sin(tt) + oris[:, [0]]
    zz = rr * np.cos(tt) + oris[:, [2]]
    yy = 0 * xx
    grid = np.stack((xx, yy, zz), axis=-1)
    return grid



"""Autograd wrapper for different backends."""

import functools

import keras

from . import _import_jax, _import_tf, _import_torch

tf = _import_tf()
jax = _import_jax()
torch = _import_torch()


class AutoGrad:
    """Wrapper class for autograd using different backends."""

    def __init__(self, verbose=False):
        self.verbose = verbose
        self.function = None

        if verbose:
            print(f"Using backend: {self.backend}")

    @property
    def backend(self):
        """Get Keras backend. Machine learning library of choice."""
        return keras.backend.backend()

    @backend.setter
    def backend(self, backend):
        """Set Keras backend. Machine learning library of choice."""
        raise ValueError("Cannot change backend currently. Needs reimport of keras.")
        # keras.config.set_backend(backend)

    def set_function(self, function):
        """Set the function to calculate the gradients of."""
        self.function = function

    def gradient(self, variable, **kwargs):
        """Returns the gradients of the function w.r.t. variable.

        Args:
            variable (Tensor): Input tensor.
            **kwargs: Keyword arguments to pass to self.function.

        Returns:
            gradients (Tensor): Gradients of the function at variable.
                ∇f(x)
        """
        variable = keras.ops.convert_to_tensor(variable)
        if self.function is None:
            raise ValueError("Function not set. Use `set_function` to set a custom function.")
        assert self.backend in [
            "torch",
            "tensorflow",
            "jax",
        ], f"Unsupported backend: {self.backend}"

        if self.backend == "torch":
            if torch is None:
                raise ImportError("PyTorch is not installed.")
            variable = variable.detach().requires_grad_(True)
            out = self.function(variable, **kwargs)
            gradients = torch.autograd.grad(out, variable)[0]
            return gradients
        elif self.backend == "tensorflow":
            if tf is None:
                raise ImportError("TensorFlow is not installed.")
            with tf.GradientTape() as tape:
                tape.watch(variable)
                out = self.function(variable, **kwargs)
            gradients = tape.gradient(out, variable)
            return gradients
        elif self.backend == "jax":
            if jax is None:
                raise ImportError("JAX is not installed.")
            func = functools.partial(self.function, **kwargs)
            return jax.grad(func)(variable)

    def gradient_and_value(self, variable, has_aux: bool = False, **kwargs):
        """Returns both the gradients w.r.t. variable and outputs of the function.

        Note that self.function should return a tuple of (out, aux) if has_aux=True.
        with aux being a tuple of auxiliary variables.
        If has_aux=False, self.function should return out only.

        Args:
            variable (Tensor): Input tensor.
            has_aux (bool): Whether the function returns auxiliary variables.
            **kwargs: Keyword arguments to pass to self.function.

        Returns:
            gradients (Tensor): Gradients of the function at variable.
                ∇f(x)
            out (Tuple or Tensor): Outputs of the function at variable.
                if has_aux: out = (f(x), aux)
                else: out = f(x)
        """
        variable = keras.ops.convert_to_tensor(variable)
        if self.function is None:
            raise ValueError("Function not set. Use `set_function` to set a custom function.")
        assert self.backend in [
            "torch",
            "tensorflow",
            "jax",
        ], f"Unsupported backend: {self.backend}"

        aux = None
        if self.backend == "torch":
            if torch is None:
                raise ImportError("PyTorch is not installed.")
            variable = variable.detach().requires_grad_(True)
            if has_aux:
                out, aux = self.function(variable, **kwargs)
            else:
                out = self.function(variable, **kwargs)
            gradients = torch.autograd.grad(out, variable)[0]
        elif self.backend == "tensorflow":
            if tf is None:
                raise ImportError("TensorFlow is not installed.")
            with tf.GradientTape() as tape:
                tape.watch(variable)
                if has_aux:
                    out, aux = self.function(variable, **kwargs)
                else:
                    out = self.function(variable, **kwargs)
            gradients = tape.gradient(out, variable)
        elif self.backend == "jax":
            if jax is None:
                raise ImportError("JAX is not installed.")
            out, gradients = jax.value_and_grad(self.function, argnums=0, has_aux=has_aux)(
                variable, **kwargs
            )
            if has_aux:
                out, aux = out
        else:
            raise ValueError(f"Unsupported backend: {self.backend}")

        if has_aux:
            return gradients, (out, aux)
        return gradients, out

    def get_gradient_jit_fn(self):
        """Returns a jitted function for calculating the gradients."""
        if self.backend == "jax":
            jax_mod = _import_jax()
            if jax_mod is None:
                raise ImportError("JAX is not installed.")
            return jax_mod.jit(self.gradient)
        elif self.backend == "tensorflow":
            tf = _import_tf()
            if tf is None:
                raise ImportError("TensorFlow is not installed.")
            return tf.function(self.gradient, jit_compile=True)
        elif self.backend == "torch":
            torch = _import_torch()
            if torch is None:
                raise ImportError("PyTorch is not installed.")
            return torch.compile(self.gradient)

    def get_gradient_and_value_jit_fn(self, has_aux: bool = False, disable_jit=False):
        """Returns a jitted function for calculating the gradients and function outputs."""
        func = lambda x, **kwargs: self.gradient_and_value(x, has_aux=has_aux, **kwargs)
        if disable_jit:
            return func
        if self.backend == "jax":
            if jax is None:
                raise ImportError("JAX is not installed.")
            return jax.jit(func)
        elif self.backend == "tensorflow":
            if tf is None:
                raise ImportError("TensorFlow is not installed.")
            return tf.function(
                func,
                jit_compile=True,
            )
        elif self.backend == "torch":
            # return torch.compile(func)
            raise NotImplementedError("Jitting not supported for torch backend.")
        else:
            raise UserWarning("You haven't set a jittable keras backend!")



"""Frequency domain ultrasound simulator.

The simulator works in the frequency domain (RFFT domain) and simulates RF data as a superposition
of scatterer responses. Every scatterer has a location and a magnitude.

To use it in your code, simply call the :func:`simulate_rf` function with the desired
transmit scheme parameters and scatterers. To simulate a sequence of multiple frames,
you can call :func:`simulate_rf` repeatedly with different scatterer positions and magnitudes
and then stack the results.

Example usage
^^^^^^^^^^^^^

A simple example of simulating RF data with a single scatterer at the center of the probe. For a
more in depth example see the notebook: :doc:`../notebooks/data/zea_simulation_example`.

.. code-block:: python

    raw_data = simulate_rf(
        scatterer_positions=np.array([[0, 0, 20e-3]]),
        scatterer_magnitudes=np.array([1.0]),
        probe_geometry=np.stack(
            [np.linspace(-20e-3, 20e-3, 64), np.zeros(64), np.zeros(64)], axis=-1
        ),
        apply_lens_correction=True,
        lens_thickness=1e-3,
        lens_sound_speed=1000,
        sound_speed=1540,
        n_ax=1024,
        center_frequency=5e6,
        sampling_frequency=20e6,
        t0_delays=np.zeros((1, 64)),
        initial_times=np.zeros(1),
        element_width=0.2e-3,
        attenuation_coef=0.5,
        tx_apodizations=np.ones((1, 64)),
    )

"""

import numpy as np
from keras import ops

from zea.beamform.lens_correction import compute_lens_corrected_travel_times


def simulate_rf(
    scatterer_positions,
    scatterer_magnitudes,
    probe_geometry,
    apply_lens_correction,
    lens_thickness,
    lens_sound_speed,
    sound_speed,
    n_ax,
    center_frequency,
    sampling_frequency,
    t0_delays,
    initial_times,
    element_width,
    attenuation_coef,
    tx_apodizations,
):
    """
    Simulates RF data for a given set of scatterers.

    Args:
        scatterer_positions (array-like): The positions of the scatterers [m] of shape (n_scat, 3).
        scatterer_magnitudes (array-like): The magnitudes of the scatterers of shape (n_scat,).
        probe_geometry (array-like): The geometry of the probe [m] of shape (n_el, 3).
        apply_lens_correction (bool): Whether to apply lens correction.
        lens_thickness (float): The thickness of the lens [m].
        lens_sound_speed (float): The speed of sound in the lens [m/s].
        sound_speed (float): The speed of sound in the medium [m/s].
        n_ax (int): The number of samples in the RF data.
        center_frequency (float): The center frequency of the pulse [Hz].
        sampling_frequency (float): The sampling frequency of the RF data [Hz].
        t0_delays (array-like): The delays of the transmitting elements [s] of shape (n_tx, n_el).
        initial_times (array-like): The initial times of the transmitting elements [s] of
            shape (n_tx,).
        element_width (float): The width of the elements [m].
        attenuation_coef (float): The attenuation coefficient [dB/cm/MHz].
        tx_apodizations (array-like): The apodizations of the transmitting elements of
            shape (n_tx, n_el).

    Returns:
        rf_data (array-like): The simulated RF data of shape (1, n_tx, n_ax, n_el, 1).
    """

    n_tx = t0_delays.shape[0]

    pulse_spectrum_fn = get_pulse_spectrum_fn(center_frequency, n_period=4)

    if not apply_lens_correction:
        dist = ops.linalg.norm(probe_geometry[None] - scatterer_positions[:, None], axis=-1)
    else:
        dist = (
            compute_lens_corrected_travel_times(
                probe_geometry,
                scatterer_positions,
                lens_thickness=lens_thickness,
                c_lens=lens_sound_speed,
                c_medium=sound_speed,
                n_iter=3,
            )
            * sound_speed
        )

    n_ax_rounded = _round_up_to_power_of_two(int(n_ax))

    freqs = (
        ops.cast(ops.arange(n_ax_rounded // 2 + 1) / n_ax_rounded, "float32") * sampling_frequency
    )

    waveform_spectrum = pulse_spectrum_fn(freqs)
    parts = []
    for tx in range(n_tx):
        tx_idx = ops.array(tx)

        # [n_scat, n_txel, rxel]
        dist_total = dist[:, None] + dist[:, :, None]

        # [n_scat, n_txel, n_rxel]
        tau_total = (
            (dist_total / sound_speed) + t0_delays[tx_idx][None, :, None] - initial_times[tx_idx]
        )

        scat_pos_relative_to_probe = scatterer_positions[:, None] - probe_geometry[None]

        # Compute 3D directivity
        theta = ops.arctan2(
            scat_pos_relative_to_probe[:, :, 0], scat_pos_relative_to_probe[:, :, 2]
        )
        phi = ops.arctan2(scat_pos_relative_to_probe[:, :, 1], scat_pos_relative_to_probe[:, :, 2])

        directivity_tx = directivity(
            freqs[None, None, None],
            theta[..., None, None],
            element_width,
            sound_speed,
        ) * directivity(
            freqs[None, None, None],
            phi[..., None, None],
            element_width,
            sound_speed,
        )
        directivity_rx = directivity(
            freqs[None, None, None],
            theta[:, None, :, None],
            element_width,
            sound_speed,
        ) * directivity(
            freqs[None, None, None],
            phi[:, None, :, None],
            element_width,
            sound_speed,
        )

        attenuation = attenuate(
            freqs[None, None, None],
            attenuation_coef=attenuation_coef,
            dist=dist_total[..., None],
        )

        spread_atten = spread(dist_total[..., None])

        result = (
            waveform_spectrum[None, None, None]
            * delay2(
                freqs[None, None, None],
                tau_total[..., None],
                n_fft=n_ax_rounded,
                sampling_frequency=sampling_frequency,
            )
            * ops.cast(
                scatterer_magnitudes[:, None, None, None]
                * tx_apodizations[tx, None, :, None, None]
                * directivity_tx
                * directivity_rx
                * attenuation
                * spread_atten,
                "complex64",
            )
        )

        # Sum over all transmitting elements and scatterers
        result = ops.sum(result, axis=[0, 1])

        result = ops.irfft((ops.real(result), ops.imag(result)))

        parts.append(result)

    rf_data = ops.stack(parts, axis=0)
    rf_data = ops.transpose(rf_data, (0, 2, 1))
    rf_data = rf_data[None, ..., None]
    rf_data = rf_data[:, :, :n_ax, :, :]
    return rf_data


def directivity(f, theta, element_width, sound_speed, rigid_baffle=True):
    """Computes the directivity of a single element.

    Args:
        f (array-like): The input frequencies [Hz].
        theta (array-like): The angles [rad].
        element_width (float): The width of the element [m].
        sound_speed (float): The speed of sound [m/s].
        rigid_baffle (bool): Whether the element is mounted on a rigid baffle,
            impacting the directivity.

    Returns:
        array-like: The directivity of the element.
    """

    wavelength = sound_speed / f

    response = sinc(element_width / wavelength * ops.sin(theta))
    if not rigid_baffle:
        response *= ops.cos(theta)
    return response


def delay2(f, tau, n_fft, sampling_frequency):
    """
    Applies a delay in the frequency domain without phase wrapping.

    Args:
        f (array-like): The input frequencies.
        tau (float): The delay to apply.
        n_fft (int): The number of samples in the FFT.
        sampling_frequency (float): The sampling frequency.

    Returns:
        array-like: The spectrum of the delay.
    """
    arg = ops.array(-1j, dtype="complex64") * ops.cast(2 * np.pi * tau * f, "complex64")
    return ops.where(
        tau < n_fft / sampling_frequency,
        ops.exp(arg),
        ops.array(0.0, dtype="complex64"),
    )


def attenuate(f, attenuation_coef, dist):
    """
    Applies attenuation to the signal in the frequency domain.

    Args:
        f (array-like): The input frequencies.
        attenuation_coef (float): The attenuation coefficient in dB/cm/MHz.
        dist (float): The distance the signal has traveled.

    Returns:
        array-like: The spectrum of the attenuation.
    """
    return ops.exp(-ops.log(10) * attenuation_coef / 20 * dist * 100 * ops.abs(f) * 1e-6)


def spread(dist, mindist=1e-4):
    """Function modeling geometric spreading of the wavefront.

    Args:
        dist (array-like): The distance the wave has traveled.
        mindist (float): The minimum distance to prevent division by zero.

    Returns:
        array-like: The geometric spreading factor of same shape as `dist`.
    """
    dist = ops.clip(dist, mindist, float("inf"))
    return mindist / dist


def hann_fd(f, width):
    """The fourier transform of a hann window in the time domain with given width."""
    denom = 1.0 - (f * width) ** 2
    num = 0.5 * sinc(f * width)
    result = num / denom
    result = ops.where(ops.abs(result) > 1.1, 0.25, result)
    return ops.nan_to_num(result, nan=0.0, posinf=0.0, neginf=0.25)


def hann_unnormalized(x, width):
    """Hann window function that is 1 at the peak. This means that the integral of the
    window function is not necessarily 1.

    Args:
        x (array-like): The input values.
        width (float): The width of the window. This is the total width from -x to x. The
            window will be nonzero in the range [-width/2, width/2].

    Returns:
        hann_vals (array-like): The values of the Hann window function.
    """
    return ops.where(ops.abs(x) < width / 2, ops.cos(np.pi * x / width) ** 2, 0)


def get_pulse_spectrum_fn(center_frequency, n_period=3.0):
    """Computes the spectrum of a sine that is windowed with a Hann window.

    Args:
        center_frequency (float): The center frequency of the pulse.
        n_period (float): The number of periods to include in the pulse.

    Returns:
        spectrum_fn (callable): A function that computes the spectrum of the pulse
        for the input frequencies in Hz.
    """
    period = n_period / center_frequency

    def spectrum_fn(f):
        return ops.array(1 / 2, "complex64") * ops.cast(
            (hann_fd(f - center_frequency, period) + hann_fd(f + center_frequency, period)),
            "complex64",
        )

    return spectrum_fn


def get_transducer_bandwidth_fn(center_frequency, bandwidth):
    """Computes the spectrum of a probe with a center frequency and bandwidth.

    Args:
        center_frequency (float): The center frequency of the probe.
        bandwidth (float): The bandwidth of the probe.

    Returns
        spectrum_fn (callable): A function that computes the spectrum of the pulse for
        the input frequencies in Hz.
    """

    def bandwidth_fn(f):
        return hann_unnormalized(ops.abs(f) - center_frequency, bandwidth)

    return bandwidth_fn


def sinc(x):
    """The normalized sinc function with a small offset to prevent division by zero."""
    x = ops.abs(np.pi * x) + 1e-9
    return ops.sin(x) / x


def _round_up_to_power_of_two(x):
    """Rounds up to the next power of two."""
    return 2 ** ops.ceil(ops.log2(x))



"""
Script to convert the EchoNet database to .npy and zea formats.
Will segment the images and convert them to polar coordinates.
"""

import os

os.environ["KERAS_BACKEND"] = "numpy"

import argparse
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path

import numpy as np
from scipy.interpolate import griddata
from tqdm import tqdm

from zea.config import Config
from zea.data import generate_zea_dataset
from zea.io_lib import load_video
from zea.utils import translate


def get_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Convert EchoNet to zea format")
    parser.add_argument(
        "--source",
        type=str,
        # path to EchoNet-Dynamic/Videos
        required=True,
    )
    parser.add_argument(
        "--output",
        type=str,
        required=True,
    )
    parser.add_argument(
        "--splits",
        type=str,
        default=None,
    )
    parser.add_argument("--output_numpy", type=str, default=None)
    parser.add_argument("--no_hyperthreading", action="store_true")
    args = parser.parse_args()
    return args


def segment(tensor, number_erasing=0, min_clip=0):
    """Segments the background of the echonet images by setting it to 0 and creating a hard edge.

    Args:
        tensor (ndarray): Input image (sc) with 3 dimensions. (N, 112, 112)
        number_erasing (float, optional): number to fill the background with.
    Returns:
        tensor (ndarray): Segmented matrix of same dimensions as input

    """
    # Start with the upper part

    # Height of the diagonal lines for the columns [0, 112]
    rows_left = np.linspace(67, 7, 61)
    rows_right = np.linspace(7, 57, 51)
    rows = np.concatenate([rows_left, rows_right], axis=0)
    for idx, row in enumerate(rows.astype(np.int32)):
        # Set everything above the edge to the number_erasing value.
        # Rows count up from 0 to 112 so row-1 is above.
        tensor[:, 0 : row - 1, idx] = number_erasing

        # Set minimum values for the edge
        if min_clip > 0:
            tensor[:, row, idx] = np.clip(tensor[:, row, idx], min_clip, 1)

    # Bottom left curve (manual fit)
    cols_left = np.linspace(0, 20, 21).astype(np.int32)
    rows_left = np.array(
        [
            102,
            103,
            103,
            104,
            104,
            105,
            105,
            106,
            106,
            107,
            107,
            107,
            108,
            108,
            109,
            109,
            109,
            110,
            110,
            111,
            111,
        ]
    )

    # Bottom right curve (manual fit)
    cols_right = np.linspace(89, 111, 23).astype(np.int32)
    rows_right = np.array(
        [
            111,
            111,
            111,
            110,
            110,
            110,
            109,
            109,
            109,
            108,
            108,
            107,
            107,
            107,
            106,
            106,
            105,
            105,
            104,
            104,
            103,
            103,
            102,
        ]
    )

    rows = np.concatenate([rows_left, rows_right], axis=0)
    cols = np.concatenate([cols_left, cols_right], axis=0)

    for row, col in zip(rows, cols):
        # Set everything under the edge to the number_erasing value.
        # Rows count up from 0 to 112 so row-1 is above.
        tensor[:, row:, col] = number_erasing
        # Set minimum values for the edge
        if min_clip > 0:
            tensor[:, row - 1, col] = np.clip(tensor[:, row - 1, col], min_clip, 1)

    return tensor


def accept_shape(tensor):
    """Acceptance algorithm that determines whether to reject an image
    based on left and right corner data.

    Args:
        tensor (ndarray): Input image (sc) with 2 dimensions. (112, 112)

    Returns:
        decision (bool): Whether or not the tensor should be rejected.

    """

    decision = True

    # Test one, check if left bottom corner is populated with values
    rows_lower = np.linspace(78, 47, 21).astype(np.int32)
    rows_upper = np.linspace(67, 47, 21).astype(np.int32)
    counter = 0
    for idx, row in enumerate(rows_lower):
        counter += np.sum(tensor[rows_upper[idx] : row, idx])

    # If it is not populated, reject the image
    if counter < 0.1:
        decision = False

    # Test two, check if the bottom right cornered with values (that are not artifacts)
    cols = np.linspace(70, 111, 42).astype(np.int32)
    rows_bot = np.linspace(17, 57, 42).astype(np.int32)
    rows_top = np.linspace(17, 80, 42).astype(np.int32)

    # List all the values
    counter = []
    for i, col in enumerate(cols):
        counter += [tensor[rows_bot[i] : rows_top[i], col]]

    flattened_counter = [float(item) for sublist in counter for item in sublist]
    # Sort and exclude the first 50 (likely artifacts)
    flattened_counter.sort(reverse=True)
    value = sum(flattened_counter[100:])

    # Reject if the baseline is too low
    if value < 5:
        decision = False

    return decision


def rotate_coordinates(data_points, degrees):
    """Function that rotates the datapoints by a certain degree.

    Args:
        data_points (ndarray): tensor containing [N,2] (x and y) datapoints.
        degrees (int): angle to rotate the datapoints with

    Returns:
       rotated_points (ndarray): the rotated data_points.

    """

    angle_radians = np.radians(degrees)
    cos_angle = np.cos(angle_radians)
    sin_angle = np.sin(angle_radians)

    rotation_matrix = np.array([[cos_angle, -sin_angle], [sin_angle, cos_angle]])
    rotated_points = rotation_matrix @ data_points.T

    return rotated_points.T


def cartesian_to_polar_matrix(
    cartesian_matrix, tip=(61, 7), r_max=107, angle=0.79, interpolation="nearest"
):
    """
    Function that converts a timeseries of a cartesian cone to a polar representation
    that is more compatible with CNN's/action selection.

    Args:
        - cartesian_matrix (2d array): (rows, cols) matrix containing time sequence
            of image_sc data.
        - tip (tuple, optional): coordinates (in indices) of the tip of the cone.
            Defaults to (61, 7).
        - r_max (int, optional): expected radius of the cone. Defaults to 107.
        - angle (float, optional): expected angle of the cone, will be used as (-angle, angle).
            Defaults to 0.79.
        - interpolation (str, optional): can be [nearest, linear, cubic]. Defaults to 'nearest'.

    Returns:
        polar_matrix (2d array): polar conversion of the input.
    """
    rows, cols = cartesian_matrix.shape
    center_x, center_y = tip

    # Create cartesian coordinates of the image data
    x = np.linspace(-center_x, cols - center_x - 1, cols)
    y = np.linspace(-center_y, rows - center_y - 1, rows)
    x, y = np.meshgrid(x, y)

    # Flatten the grid and values
    data_points = np.column_stack((x.ravel(), y.ravel()))
    data_points = rotate_coordinates(data_points, -90)
    data_values = cartesian_matrix.ravel()

    # Define new points to sample from in the region of the data.
    # R_max and Theta are found manually. R_max differs from the number of rows in EchoNet!
    r = np.linspace(0, r_max, rows)
    theta = np.linspace(-angle, angle, cols)
    r, theta = np.meshgrid(r, theta)

    x_polar = r * np.cos(theta)
    y_polar = r * np.sin(theta)
    new_points = np.column_stack((x_polar.ravel(), y_polar.ravel()))

    # Interpolate and reshape to 2D matrix
    polar_values = griddata(
        data_points, data_values, new_points, method=interpolation, fill_value=0
    )
    polar_matrix = np.rot90(polar_values.reshape(cols, rows), k=-1)
    return polar_matrix


def find_split_for_file(file_dict, target_file):
    """Function that finds the split for a given file in a dictionary."""
    for split, files in file_dict.items():
        if target_file in files:
            return split
    return "rejected"


class H5Processor:
    """
    Stores a few variables and paths to allow for hyperthreading.
    """

    def __init__(
        self,
        path_out_h5,
        path_out=None,
        num_val=500,
        num_test=500,
        range_from=(0, 255),
        range_to=(-60, 0),
        splits=None,
    ):
        self.path_out_h5 = Path(path_out_h5)
        self.path_out = Path(path_out) if path_out else None
        self.num_val = num_val
        self.num_test = num_test
        self.range_from = range_from
        self.range_to = range_to
        self.splits = splits
        self._process_range = (0, 1)

        # Ensure train, val, test, rejected paths exist
        for folder in ["train", "val", "test", "rejected"]:
            if self._to_numpy:
                (self.path_out / folder).mkdir(parents=True, exist_ok=True)
            (self.path_out_h5 / folder).mkdir(parents=True, exist_ok=True)

    @property
    def _to_numpy(self):
        return self.path_out is not None

    def translate(self, data):
        """Translate the data from the processing range to final range."""
        return translate(data, self._process_range, self.range_to)

    def get_split(self, hdf5_file: str, sequence):
        """Determine the split for a given file."""
        # Always check acceptance
        accepted = accept_shape(sequence[0])

        # Previous split
        if self.splits is not None:
            split = find_split_for_file(self.splits, hdf5_file)
            assert accepted == (split != "rejected"), "Rejection mismatch"
            return split

        # New split
        if not accepted:
            return "rejected"

        # This inefficient counter works with hyperthreading
        # TODO: but it is not reproducible!
        val_counter = len(list((self.path_out_h5 / "val").iterdir()))
        test_counter = len(list((self.path_out_h5 / "test").iterdir()))

        # Determine the split
        if val_counter < self.num_val:
            return "val"
        elif test_counter < self.num_test:
            return "test"
        else:
            return "train"

    def __call__(self, avi_file):
        """
        Processes a single h5 file using the class variables and the filename given.
        """
        hdf5_file = avi_file.stem + ".hdf5"
        sequence = load_video(avi_file)

        assert sequence.min() >= self.range_from[0], f"{sequence.min()} < {self.range_from[0]}"
        assert sequence.max() <= self.range_from[1], f"{sequence.max()} > {self.range_from[1]}"

        # Translate to [0, 1]
        sequence = translate(sequence, self.range_from, self._process_range)

        sequence = segment(sequence, number_erasing=0, min_clip=0)

        split = self.get_split(hdf5_file, sequence)
        accepted = split != "rejected"

        out_h5 = self.path_out_h5 / split / hdf5_file
        if self._to_numpy:
            out_dir = self.path_out / split / avi_file.stem
            out_dir.mkdir(parents=True, exist_ok=True)

        polar_im_set = []
        for i, im in enumerate(sequence):
            if self._to_numpy:
                np.save(out_dir / f"sc{str(i).zfill(3)}.npy", im)

            if not accepted:
                continue

            polar_im = cartesian_to_polar_matrix(im, interpolation="cubic")
            polar_im = np.clip(polar_im, *self._process_range)
            if self._to_numpy:
                np.save(
                    out_dir / f"polar{str(i).zfill(3)}.npy",
                    polar_im,
                )
            polar_im_set.append(polar_im)

        if accepted:
            polar_im_set = np.stack(polar_im_set, axis=0)

        # Check the ranges
        assert sequence.min() >= self._process_range[0], sequence.min()
        assert sequence.max() <= self._process_range[1], sequence.max()

        zea_dataset = {
            "path": out_h5,
            "image_sc": self.translate(sequence),
            "probe_name": "generic",
            "description": "EchoNet dataset converted to zea format",
        }
        if accepted:
            zea_dataset["image"] = self.translate(polar_im_set)
        return generate_zea_dataset(**zea_dataset)


if __name__ == "__main__":
    args = get_args()

    if args.splits is not None:
        # Reproduce a previous split...
        split_yaml_dir = Path(args.splits)
        splits = {"train": None, "val": None, "test": None}
        for split in splits:
            yaml_file = split_yaml_dir / (split + ".yaml")
            assert yaml_file.exists(), f"File {yaml_file} does not exist."
            splits[split] = Config.from_yaml(yaml_file)["file_paths"]
    else:
        splits = None

    # List the files that have an entry in path_out_h5 already
    files_done = []
    for _, _, filenames in os.walk(args.output):
        for filename in filenames:
            files_done.append(filename.replace(".hdf5", ""))

    # List all files of echonet and exclude those already processed
    path_in = Path(args.source)
    h5_files = path_in.glob("*.avi")
    h5_files = [file for file in h5_files if file.stem not in files_done]
    print(f"Files left to process: {len(h5_files)}")

    # Run the processor
    processor = H5Processor(path_out_h5=args.output, path_out=args.output_numpy, splits=splits)

    print("Starting the conversion process.")

    if not args.no_hyperthreading:
        with ProcessPoolExecutor() as executor:
            futures = {executor.submit(processor, file): file for file in h5_files}
            for future in tqdm(as_completed(futures), total=len(h5_files)):
                future.result()
    else:
        for file in tqdm(h5_files):
            processor(file)

    print("All tasks are completed.")



"""
Tensorflow utilities
"""

import keras
import numpy as np
import tensorflow as tf
from keras.layers import Lambda, LeakyReLU, ReLU

PI = tf.experimental.numpy.pi


def antirect(x):
    """Function that implements the antirectifier activation"""
    mean, _ = tf.nn.moments(x, axes=-1, keepdims=True)
    x = tf.math.l2_normalize(x - mean, axis=-1)
    return tf.nn.crelu(x)


def get_activation(activation: str = None):
    """Get activation function given string.

    Args:
        activation (str, optional): name of activation
            function. Defaults to None.

    Raises:
        ValueError: Cannot find activation function

    Returns:
        Tensorflow activation function
    """
    if activation.lower() == "relu":
        return ReLU()
    elif activation.lower() == "leakyrelu":
        return LeakyReLU()
    elif activation.lower() == "swish":
        return Lambda(lambda x: keras.activations.swish(x))
    elif activation.lower() == "sigmoid":
        return Lambda(lambda x: keras.activations.sigmoid(x))
    elif activation is None:
        return Lambda(lambda x: x)
    else:
        raise ValueError("Unknown activation function.")


def tf_cropping_and_padding(input_shape, target_shape):
    """Crop or pad a tensor to the specified shape.

    Args:
        input_shape: A list or tuple of integers representing the input shape.
        target_shape: A list or tuple of integers representing the desired shape.

    Returns:
        A tensorflow cropping layer (2D) for 4D tensors.
    """
    assert len(input_shape) == len(target_shape) == 2, "can only do 2D cropping"

    # Calculate the amount of cropping needed for each dimension
    diff = np.array(input_shape) - np.array(target_shape)
    cropping = ((0, diff[0]), (0, diff[1]))

    # replace negative values with zero and make padding tuple of tuples
    padding = tuple(tuple(0 if x > 0 else -x for x in row) for row in cropping)
    cropping = tuple(tuple(0 if x < 0 else x for x in row) for row in cropping)

    # Create a Cropping2D layer with the calculated crop amounts
    cropping_layer = keras.layers.Cropping2D(cropping=cropping)

    # Create a ZeroPadding2D layer with the calculated padding amounts
    padding_layer = keras.layers.ZeroPadding2D(padding=padding)

    return cropping_layer, padding_layer


def tf_complex_resize(tensor, image_size):
    """Resize / interpolate complex tensor"""
    magnitude = tf.math.abs(tensor)
    phase = tf.math.angle(tensor)

    # unwrap phase
    phase = tf_unwrap(phase)

    magnitude = tf.image.resize(magnitude, image_size)
    phase = tf.image.resize(phase, image_size)

    phase = tf.complex(tf.cast(0.0, dtype=phase.dtype), phase)
    magnitude = tf.cast(magnitude, dtype=tf.complex64)

    # complex = magnitude * exp(1j * phase)
    return tf.math.multiply(magnitude, tf.math.exp(phase))


def tf_unwrap(tensor, axis=0):
    """Tensorflow phase unwrapping function"""
    pi_tf = tf.cast(PI, dtype=tensor.dtype)
    dphi = tf.experimental.numpy.diff(tensor, axis=axis)
    dphi_pad = tf.gather(dphi, 0, axis=axis)
    dphi_pad = tf.expand_dims(dphi_pad, axis=axis)
    dphi = tf.concat([dphi_pad, dphi], axis=axis)
    dphi_m = ((dphi + pi_tf) % (2.0 * pi_tf)) - pi_tf
    dphi_m = tf.where((dphi_m == -PI) & (dphi > 0.0), pi_tf, dphi_m)
    phi_adj = dphi_m - dphi
    phi_adj = tf.where(tf.abs(dphi) < pi_tf, 0.0, phi_adj)
    return tensor + tf.math.cumsum(phi_adj, axis=axis)



"""
Apodization functions
"""

import numpy as np
from scipy import signal


def square_wave_apodization(n_el: int, block_size: float):
    """Returns a square wave apodization.

    Used for incoherent beamforming.

    Args:
        n_el (int): Total number of elements in array.
        block_size (float): In terms of number of elements that will be
            high/low. Can be a float.

    Returns:
        apod (ndarray): array of size n_el.

    Example:

        +1 +1 +1 +1             +1 +1 +1 +1
                    -1 -1 -1 -1

        <----------> block_size = 4, n_el = 12
    """

    lambd = 2 * block_size / n_el
    freq = 1 / lambd
    t = np.linspace(0, 1, n_el)
    apod = signal.square(2 * np.pi * freq * t, duty=0.5)

    # weirdly last element is flipped
    apod[-1] = -1 * apod[-1]
    return apod
